import os
import sys
import subprocess
from hashlib import md5
from itertools import chain

from app.drivers.options import psmtable_options, percolator_options
from app.drivers.base import PSMDriver

from app.readers import tsv as tsvreader
from app.readers import mzidplus as mzidreader
from app.dataformats import mzidtsv as psmhead
from app.dataformats.prottable import HEADER_NO_FULLQ_PSMS

from app.actions.psmtable import splitmerge as splitmerge
from app.actions.psmtable import refine as refine
from app.actions.psmtable import percolator as perco
from app.actions.psmtable import filtering
from app.actions.psmtable import isosummarize

from app.dataformats import prottable as prottabledata

from app.writers import tsv as writer


class TSVConcatenateDriver(PSMDriver):
    """Concatenates TSVs, takes care of headers"""
    outsuffix = '_concat.tsv'
    command = 'concat'
    commandhelp = ('Merges multiple TSV tables of MSGF+ output.'
                   'Make sure headers are same in all files.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['multifiles'],
                                                psmtable_options))

    def prepare(self):
        # do not read PSMs, multiple files passed and they will be checked
        # if headers matched
        self.first_infile = self.fn[0]
        self.oldheader = tsvreader.get_tsv_header(self.first_infile)

    def set_features(self):
        self.header = self.oldheader
        self.psms = splitmerge.merge_mzidtsvs(self.fn, self.header)


class TSVSplitDriver(PSMDriver):
    """Splits MSGF PSM table on contents of certain column. Each
    row in file is piped to an output file. Which output files
    the row is written to depends on the contents of the selected
    column"""
    outsuffix = '_split.tsv'
    command = 'split'
    commandhelp = 'Splits an MSGF TSV PSM table into multiple new tables'

    def set_options(self):
        super().set_options()
        options = self.define_options(['splitcol'], psmtable_options)
        self.options.update(options)

    def set_features(self):
        self.header = self.oldheader[:]
        splitfield = splitmerge.get_splitfield(self.oldheader, self.splitcol)
        self.psms = splitmerge.generate_psms_split(self.oldpsms, splitfield)

    def write(self):
        base_outfile = os.path.join(self.outdir, '{}.tsv')
        writer.write_multi_mzidtsv(self.header, self.oldheader, self.psms,
                                    base_outfile)


class Perco2PSMDriver(PSMDriver):
    """
    Adds percolator data from mzid file to table.
    """
    outsuffix = '_fdr.tsv'
    command = 'perco2psm'
    commandhelp = (
            'Calculates FDR from percolator output and adds FDR and percolator '
            ' to the corresponding TSV PSM tables. FDR calculation method is TD-'
            'competition.'
            )

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['multifiles', 'mzidfns', 'percofn',
            'filtpep', 'filtpsm', 'qvalitypsms', 'qvalitypeps'], psmtable_options))

    def prepare(self):
        # multiple PSM tables passed so do not read here, match with mzid
        pass

    def set_features(self):
        if self.qvalitypsms:
            qvpsmhead = tsvreader.get_tsv_header(self.qvalitypsms)
            qvpephead = tsvreader.get_tsv_header(self.qvalitypeps)
            qvpsms = tsvreader.generate_split_tsv_lines(self.qvalitypsms, qvpsmhead)
            qvpeps = tsvreader.generate_split_tsv_lines(self.qvalitypeps, qvpephead)
            self.percopsms = perco.get_scores_fdrs_from_percoqvality(self.percofn, qvpsms, qvpeps)
        else:
            self.percopsms = perco.get_scores_fdrs_from_percoxml(self.percofn)

                
    def write(self):
        for psmfn, mzidfn in zip(self.fn, self.mzidfns):
            oldheader = tsvreader.get_tsv_header(psmfn)
            header = perco.get_header_with_percolator(oldheader)
            outfn = self.create_outfilepath(psmfn, self.outsuffix)
            mzns = mzidreader.get_mzid_namespace(mzidfn)
            mzidsr = mzidreader.mzid_spec_result_generator(mzidfn, mzns)
            psms = tsvreader.generate_split_tsv_lines(psmfn, oldheader)
            psms_perco = perco.add_fdr_to_mzidtsv(psms, mzidsr, mzns,
                    self.percopsms)
            if self.filtpsm:
                psms_perco = filtering.filter_psms_conf(psms_perco, psmhead.HEADER_PSMQ,
                        self.filtpsm, True)
            if self.filtpep:
                psms_perco = filtering.filter_psms_conf(psms_perco, psmhead.HEADER_PEPTIDE_Q,
                        self.filtpep, True)
            writer.write_tsv(header, psms_perco, outfn)


class ConfidenceFilterDriver(PSMDriver):
    outsuffix = '_filtconf.txt'
    command = 'conffilt'
    commandhelp = 'Filters PSMs by their confidence level. '

    def set_options(self):
        super().set_options()
        options = self.define_options(['confcol', 'confpattern', 'conflvl',
                                       'conftype', 'unroll'], psmtable_options)
        self.options.update(options)

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.lowerbetter = self.conftype == 'lower'

    def set_features(self):
        self.header = self.oldheader[:]
        if self.confpattern:
            confkey = tsvreader.get_cols_in_file(self.confpattern,
                                                 self.header, True)
        elif self.confcol:
            confkey = self.header[int(self.confcol) - 1]
        else:
            print('Must define either --confcol or --confcolpattern')
            sys.exit(1)
        self.psms = filtering.filter_psms_conf(self.oldpsms,
                                       confkey,
                                       self.conflvl,
                                       self.lowerbetter)


class DuplicateFilterDriver(PSMDriver):
    outsuffix = '_dedup.txt'
    command = 'deduppsms'
    commandhelp = 'Removes duplicate PSMs by file/scan/sequence combination'

    def set_options(self):
        super().set_options()
        options = self.define_options(['spectracol', 'peptidecolpattern'], psmtable_options)
        self.options.update(options)

    def set_features(self):
        self.header = self.oldheader[:]
        specfncolnr = int(self.spectracol) - 1
        specfncol = self.oldheader[specfncolnr]
        if self.peptidecolpattern:
            seqcol = tsvreader.get_columns_by_pattern(self.oldheader, self.peptidecolpattern)
            if len(seqcol) > 1:
                print('Your pattern to find the sequence colum in the PSM table resulted '
                        f'in multiple column names returned: [{", ".join(seqcol)}]')
                sys.exit(1)
            else:
                seqcol = seqcol[0]
        else:
            seqcol = psmhead.HEADER_PEPTIDE
        scancol = psmhead.HEADER_SCANNR
        self.psms = filtering.remove_duplicate_psms(self.oldpsms, specfncol, scancol, seqcol)



class SequenceFilterDriver(PSMDriver):
    outsuffix = '_filtseq.txt'
    command = 'seqfilt'
    commandhelp = 'Filters PSMs by their sequence'
    lookuptype = 'searchspace'

    def set_options(self):
        super().set_options()
        options = self.define_options(['fullprotein', 'lookupfn', 'unroll',
            'minlength'], psmtable_options)
        self.options.update(options)
        options = self.define_options(['deamidate', 'forcetryp', 'insourcefrag'],
                percolator_options)
        self.options.update(options)

    def set_features(self):
        self.header = self.oldheader[:]
        if self.fullprotein:
            self.psms = filtering.filter_whole_proteins(self.oldpsms, 
                    self.lookup, psmhead.HEADER_PEPTIDE, self.deamidate, 
                    self.minlength, self.forcetryp)
        else:
            self.psms = filtering.filter_known_searchspace(self.oldpsms, 
                    self.lookup, psmhead.HEADER_PEPTIDE, self.insourcefrag, self.deamidate)


class SequenceMatchDriver(PSMDriver):
    outsuffix = '_seqmatch.txt'
    command = 'seqmatch'
    commandhelp = 'Adds column to PSMs with matched sequences'
    lookuptype = 'searchspace'

    def set_options(self):
        super().set_options()
        options = self.define_options(['fullprotein', 'lookupfn', 'unroll', 'minlength', 'matchfilename'],
                psmtable_options)
        self.options.update(options)
        options = self.define_options(['deamidate', 'forcetryp', 'insourcefrag'], percolator_options)
        self.options.update(options)

    def set_features(self):
        self.header = [*self.oldheader[:], self.matchfilename]
        self.psms = filtering.match_sequence(self.oldpsms, self.lookup, psmhead.HEADER_PEPTIDE,
                self.matchfilename, self.insourcefrag, self.deamidate, self.minlength, self.forcetryp)


class PSMTableRefineDriver(PSMDriver):
    # gene, quant, pg, spectra
    outsuffix = '_refined.txt'
    lookuptype = 'psm'
    command = 'psmtable'
    commandhelp = ('Add columns to mzidtsv with genes, quantification, protein groups, '
                   'etc. which are stored in a lookup specified with --dbfile')

    def set_options(self):
        super().set_options()
        options = self.define_options(['oldpsmfile', 'lookupfn', 'precursor', 'isobaric',
            'minpurity', 'unroll', 'spectracol', 'addbioset', 'addmiscleav', 'genes',
            'proteingroup', 'fasta', 'genefield', 'fastadelim'], psmtable_options)
        self.options.update(options)

    def set_features(self):
        """Creates iterator to write to new tsv. Contains input tsv
        lines plus quant data for these."""
        # First prepare the data, read PSM table to SQLite
        specfncolnr = int(self.spectracol) - 1
        specfncol = self.oldheader[specfncolnr]
        fastadelim, genefield = self.get_fastadelim_genefield(self.fastadelim,
                                                              self.genefield)

        if self.fasta:
            fasta_md5 = refine.get_fasta_md5(self.fasta)
        else:
            fasta_md5 = False

        # If appending to previously refined PSM table, reuse DB and shift rows
        if self.oldpsmfile:
            oldfasta_md5 = self.lookup.get_fasta_md5()
            if fasta_md5 != oldfasta_md5:
                print('WARNING, FASTA database used in old PSM table differs '
                        'from the passed database (or this cannot be determined '
                        'due to version differences), this may cause problems, as '
                        'msstitch will use the old database for PSM annotation.')
            shiftrows = self.lookup.get_highest_rownr() + 1
            proteins = set([x for x in self.lookup.get_protids()])
            self.lookup.drop_psm_indices()
        else:
            shiftrows, oldfasta_md5 = 0, False

        if self.proteingroup:
            if not fasta_md5 and not oldfasta_md5:
                # In case of old Fasta already stored it will be fine to protein group
                print('Cannot create protein group without supplying FASTA search '
                        'database file')
                sys.exit(1)
            self.tabletypes.append('proteingroup')
            self.lookup.drop_pgroup_tables()
        self.lookup.add_tables(self.tabletypes)

        # Need to place this here since we cannot store before having done add tables, but that
        # has to be done after getting proteingroup knowledge, which depends on knowledge of 
        # having passed an oldpsmfile (because of oldfasta_md5):
        if not self.oldpsmfile:
            proteins = refine.store_proteins_descriptions(self.lookup, self.fasta,
                    fasta_md5, self.fn, self.oldheader, fastadelim, genefield)

        refine.create_psm_lookup(self.fn, self.oldheader, proteins, self.lookup, 
                shiftrows, self.unroll, specfncol, fastadelim, genefield)
        isob_header = [x[0] for x in self.lookup.get_all_quantmaps()] if self.isobaric else False
        self.header = refine.create_header(self.oldheader, self.genes, 
                self.proteingroup, self.precursor, isob_header, self.addbioset, 
                self.addmiscleav, specfncolnr)
        psms = self.oldpsms
        # Now pass PSMs through multiple generators to add info
        if self.genes:
            psms = refine.add_genes_to_psm_table(psms, self.lookup)
        if self.isobaric or self.precursor:
            psms = refine.generate_psms_quanted(self.lookup, shiftrows, psms,
                    isob_header, self.isobaric, self.precursor, self.min_purity)
        psms = refine.generate_psms_spectradata(self.lookup, shiftrows, 
                psms, self.addbioset, self.addmiscleav)
        if self.oldpsmfile:
            prevheader = tsvreader.get_tsv_header(self.oldpsmfile)
            previouspsms = tsvreader.generate_split_tsv_lines(self.oldpsmfile, prevheader)
            psms = chain(previouspsms, psms)
        # Enforce proteingroup last, since it has to come AFTER the chaining of old + new PSMs
        # In theory you could do it before, but that makes no sense since in a big experiment you
        # also do not map PSMs to genes differently? If that is needed, you have to run multiple
        # experiments.
        if self.proteingroup:
            refine.build_proteingroup_db(self.lookup)
            psms = refine.generate_psms_with_proteingroups(psms, self.lookup, specfncol, self.unroll)
        self.psms = psms
        

class IsoSummarizeDriver(PSMDriver):
    outsuffix = '_ratio_isobaric.txt'
    command = 'isosummarize'
    commandhelp = ('Produce isobaric summarized data from a '
                   'PSM table containing raw intensities.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['quantcolpattern', 'denompatterns',
            'denomcols', 'mediansweep', 'medianintensity', 'median_or_avg',
            'keep_psms_na', 'minint', 'featcol', 'logisoquant', 'splitmultientries',
            'mediannormalize', 'mednorm_factors'], psmtable_options))

    def set_features(self):
        denomcols = False
        if self.denomcols is not None:
            denomcols = [self.number_to_headerfield(col, self.oldheader)
                         for col in self.denomcols]
        elif self.denompatterns is not None:
            denomcolnrs = [tsvreader.get_columns_by_combined_patterns(
                self.oldheader, [self.quantcolpattern, pattern]) for pattern in self.denompatterns]
            denomcols = set([col for cols in denomcolnrs for col in cols])
        elif not self.mediansweep and not self.medianintensity:
            raise RuntimeError('Must define either denominator column numbers '
                               'or regex pattterns to find them')
        quantcols = tsvreader.get_columns_by_pattern(self.oldheader,
                                               self.quantcolpattern)
        mn_factors = False
        if self.mednorm_factors:
            mnhead = tsvreader.get_tsv_header(self.mednorm_factors)
            mn_factors = tsvreader.generate_split_tsv_lines(self.mednorm_factors, mnhead)
        nopsms = [isosummarize.get_no_psms_field(qf) for qf in quantcols]
        if self.featcol:
            self.get_column_header_for_number(['featcol'], self.oldheader)
            self.header = [self.featcol] + quantcols + nopsms + [HEADER_NO_FULLQ_PSMS]
        else:
            self.header = (self.oldheader +
                           ['ratio_{}'.format(x) for x in quantcols])
        self.psms = isosummarize.get_isobaric_ratios(self.fn, self.oldheader,
                quantcols, denomcols, self.mediansweep, self.medianintensity,
                self.median_or_avg, self.minint, False, False, self.featcol,
                False, False, False, self.logisoquant, self.mediannormalize,
                mn_factors, self.keepnapsms, self.splitmultientries)


class DeleteSetDriver(PSMDriver):
    outsuffix = '_deletedset.txt'
    command = 'deletesets'
    lookuptype = 'psm'
    commandhelp = """Remove sample sets from an existing PSM table and its lookup file"""

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lookupfn', 'setnames'], psmtable_options))
        self.options['lookupfn'].update({'required': False, 'default': None})

    def set_features(self):
        # In rare cases, a lookup is not used and we just filter the PSM table. I.e when
        # making a new lookup from a new+old PSM table.
        if self.lookup:
            self.lookup.delete_sample_set_shift_rows(self.setnames)
        self.header = self.oldheader
        self.psms = filtering.filter_psms_remove_set(self.oldpsms, self.setnames)


class LuciphorDriver(PSMDriver):
    outsuffix = '_lucxor.txt'
    command = 'luciphor'
    commandhelp = """Run luciphor on PSM table (of a single set) to get a table 
    with PSMs with PTMs, containing false-localization rates and Luciphor scoring
    """

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lucitemplate', 'mods', 'luciptms',
            'ms2tolerance', 'luciminpsms', 'luci_frag', 'maxlength', 'maxcharge',
            'mzmlpath', 'lucimemory', 'lucithread'], psmtable_options))

    def set_features(self):
        # parse passed mods and get masses etc from DB
        luciptms = {}
        for lptm in self.luciptms:
            lptm.split(':')
            aa = lptm[1] if len(lptm) == 2 else ''
            luciptms[lptm[0]] = aa
        fixm, varm, ptms = self.lookup.get_mods_luciphor(luciptms)

        # parse MS2 tolerance
        if self.ms2tol[-2:] == 'Da':
            toltype = 0
            mstol = self.ms2tol[:-2]
        elif self.ms2tol[-3:] == 'ppm':
            toltype = 1
            mstol = self.ms2tol[:-3]
        else:
            print('If defining MS2 tolerance, use formats --ms2tol 10ppm or --ms2tol 0.02Da')
            sys.exit(1)
        try:
            float(mstol)
        except TypeError:
            print(f'MS2 tolerance value {mstol} should be a number!')
            sys.exit(1)
        ptmmods = {}
        
        # 
        algo = {'hcd': 1, 'cid': 0}[self.luci_frag.lower()]
        # Run in block with cleanup
        try:
            luci_in = create_luciphor_input(self.oldpsms, lucitemplate, self.mods, self.luciptms, mstol, toltype, self.mzmlpath, algo, self.maxcharge, self.maxlength, self.luci_minpsms, self.lucithread)
            self.write_luci_input(luci_in)
            subprocess.run(['luciphor2', f'-Xmx{self.lucimem}G', '.luciphor_config.txt'])
            self.psms = parse_luciphor_output(luciptms, luciscores, ptmpsms, tdb, self.luciminscore,
                    [*varm, *ptms])
        except Exception:
            raise
        finally:
            pass 
            # FIXME cleanup all_scores, luciphor.out, etc etc

    def write_luci_input(self, luci_inputs):
        luci_in_psm_header = ('srcFile', 'scanNum', 'charge', 'PSMscore', 'peptide', 'modSites')
        config_contents = next(luci_inputs)
        with open('luciphor_config_input.txt', 'w') as wfp:
            wfp.write(config_contents)
        writer.write_tsv(luci_in_psm_header, luci_inputs, 'luci_input')
                


    #with open('luciphor_psm_input', 'w') as wfp:
#        wfp.write('srcFile\tscanNum\tcharge\tPSMscore\tpeptide\tmodSites')
#            wfp.write('\n{}\t{}\t{}\t{}\t{}\t{}'.format(psm[dm.HEADER_SPECFILE],
#                psm[dm.HEADER_SCANNR], psm[dm.HEADER_CHARGE], psm[dm.HEADER_PSMQ],
#                pep, ','.join(['{}={}'.format(x[0], x[1]) for x in sites])))
        #writer.

        pass

        # need to store PTM/PSMs in the output as well
        ## need some try/finally clause to clean up tmp files (luci_input)
        ## need a reader for the scorefile, luciphor.out file


def create_mods():
    # FIXME this func to actions
    for mod in varmods:
        realmodmass = mod[0]
        mmass, mres, mfm, mprotpos, mname = mod
        mp = modpos(mod)
        if mp in fixedmods:
            nonblocked_fixed = NON_BLOCKING_MODS[mname] if mname in NON_BLOCKING_MODS else []
            blocked_fixmass = sum([float(x[0]) for x in fixedmods[mp] if x[-1] not in nonblocked_fixed])
            mod[0] = str(round(-(blocked_fixmass - float(mod[0])), 5))



import os
from app.readers import xml as basereader
from app.readers import xmlformatting as formatting


def mzmlfn_ms2_spectra_generator(mzmlfiles):
    for fn, spec, ns in mzmlfn_spectra_generator(mzmlfiles):
        specparams = spec.findall('{%s}cvParam' % ns['xmlns'])
        mslvl = fetch_cvparam_value_by_name(specparams, 'ms level')
        if mslvl != '2':
            continue
        specscanid = spec.attrib['id']
        scan = basereader.find_element_xpath(spec, 'scan', ns)
        rtcv = [x for x in scan.findall('{%s}cvParam' % ns['xmlns']) if x.attrib['accession'] == 'MS:1000016'][0]
        if rtcv.attrib['unitAccession'] == 'UO:0000031':
            rt = float(rtcv.attrib['value'])
        elif rtcv.attrib['unitAccession'] == 'UO:0000010':
            rt = float(rtcv.attrib['value']) / 60
        iit, ionmob  = fetch_cvparams_values_from_subel(spec, 'scan', [
            'ion injection time', 
            'inverse reduced ion mobility'], ns)
        mz, charge = fetch_cvparams_values_from_subel(spec, 'selectedIon', [
            'selected ion m/z',
            'charge state'], ns)
        yield fn, {'specscanid': specscanid, 'ionmob': ionmob, 'rt': rt, 'iit': iit, 
                'mz': mz, 'charge': charge}
        formatting.clear_el(spec)
        # FIXME generator clears since 2.14 so this may be removed
        # or we revert since it broke at least one external script 


def mzmlfn_spectra_generator(mzmlfiles):
    for fn in mzmlfiles:
        ns = basereader.get_namespace(fn)
        spectra = basereader.generate_xmltags(
            fn,
            'spectrum',
            ['offset',
             ],
            ns)
        for spectrum in spectra:
            yield os.path.basename(fn), spectrum, ns


def fetch_cvparams_values_from_subel(base, subelname, paramnames, ns):
    """Searches a base element for subelement by name, then takes the
    cvParams of that subelement and returns the values as a list
    for the paramnames that match. Value order in list equals input
    paramnames order."""
    sub_el = basereader.find_element_xpath(base, subelname, ns)
    cvparams = sub_el.findall('{%s}cvParam' % ns['xmlns'])
    output = []
    for param in paramnames:
        output.append(fetch_cvparam_value_by_name(cvparams, param))
    return output


def fetch_cvparam_value_by_name(params, name):
    for cvparam in params:
        try:
            if cvparam.attrib['name'] == name:
                return cvparam.attrib['value']
        except KeyError:
            pass
    return False



HEADER_PROTEIN = 'Protein ID'
HEADER_PROTEINS = 'Protein ID(s)'
HEADER_GENEID = 'Gene ID'
HEADER_GENENAME = 'Gene Name'
HEADER_DESCRIPTION = 'Description'
HEADER_COVERAGE = 'Coverage'
HEADER_NO_PROTEIN = 'Protein count'
HEADER_CONTENTPROT = 'Proteins in group'
HEADER_NO_UNIPEP = 'Unique peptide count'
HEADER_NO_PEPTIDE = 'Peptide count'
HEADER_NO_PSM = 'PSM count'
HEADER_NO_PSMS_SUFFIX = ' - Quanted PSM count'
HEADER_NO_FULLQ_PSMS = 'Fully quanted PSM count'
HEADER_AREA = 'MS1 precursor area'
HEADER_QVAL = 'q-value'
HEADER_PEP = 'PEP'
HEADER_QVAL_MODELED = 'q-value (linear model)'
HEADER_QSCORE = 'Q-score best peptide'

PICKED_HEADER = [HEADER_QSCORE, HEADER_QVAL]
ACCESSIONS = {
        'prottable': HEADER_PROTEIN, 
        'genetable': HEADER_GENEID,
        'associdtable': HEADER_GENENAME,
        }
TPROT_HEADER_ACCS = [HEADER_PROTEIN, HEADER_GENENAME, HEADER_GENEID, HEADER_PROTEIN]



import re
from hashlib import md5
from collections import OrderedDict

from app.readers import tsv as tsvreader
from app.readers import fasta as fastareader
from app.dataformats import mzidtsv as mzidtsvdata

from app.lookups.sqlite import psms as lookups

DB_STORE_CHUNK = 100000


def create_header(oldheader, genes, proteingroup, precursor, isob_header, bioset,
        miscleav, specfncolnr):
    header = oldheader[:]
    p_ix = header.index(mzidtsvdata.HEADER_PROTEIN) + 1
    if genes:
        newfields = [mzidtsvdata.HEADER_GENE, mzidtsvdata.HEADER_SYMBOL,
                     mzidtsvdata.HEADER_DESCRIPTION]
        header = header[:p_ix] + newfields + oldheader[p_ix:]
    if proteingroup:
        header = header[:p_ix] + mzidtsvdata.HEADER_PG + header[p_ix:]
    if precursor:
        header += [mzidtsvdata.HEADER_PRECURSOR_QUANT, mzidtsvdata.HEADER_PRECURSOR_FWHM]
    if isob_header:
        header += [mzidtsvdata.HEADER_PREC_PURITY, *isob_header]
    psmdatafields = mzidtsvdata.MOREDATA_HEADER
    if bioset:
        psmdatafields = [mzidtsvdata.HEADER_SETNAME] + psmdatafields
    if miscleav:
        psmdatafields.append(mzidtsvdata.HEADER_MISSED_CLEAVAGE)
    header = header[:specfncolnr +1] + [x for x in psmdatafields if x not in header] + header[specfncolnr + 1:]
    return header


def create_psm_lookup(fn, header, proteins, pgdb, shiftrows, unroll, specfncol, fastadelim, genefield):
    """Reads PSMs from file, stores them to a database backend in chunked PSMs.
    """
    mzmlmap = pgdb.get_mzmlfile_map()
    sequences = {}
    for psm in tsvreader.generate_split_tsv_lines(fn, header):
        seq = tsvreader.get_psm_sequence(psm, unroll)
        sequences[seq] = 1
    pepseqmap = pgdb.get_peptide_seq_map()
    pgdb.store_pepseqs(((seq,) for seq in sequences if seq not in pepseqmap))
    pepseqmap = pgdb.get_peptide_seq_map()
    psms = []
    for row, psm in enumerate(tsvreader.generate_split_tsv_lines(fn, header)):
        row += shiftrows
        specfn, psm_id, specscanid, seq, score = tsvreader.get_psm(psm, unroll, specfncol)
        if len(psms) % DB_STORE_CHUNK == 0:
            pgdb.store_psms(psms)
            psms = []
        psms.append({'rownr': row,
                     'psm_id': psm_id,
                     'seq': pepseqmap[seq],
                     'score': score,
                     'specfn': mzmlmap[specfn],
                     'spec_id': '{}_{}'.format(mzmlmap[specfn], specscanid),
                     })
    pgdb.store_psms(psms)
    pgdb.index_psms()
    store_psm_protein_relations(fn, header, pgdb, proteins, specfncol)


def get_fasta_md5(fastafn):
    fasta_md5 = md5()
    with open(fastafn, 'rb') as fp:
        for chunk in iter(lambda: fp.read(4096), b''):
            fasta_md5.update(chunk)
    return fasta_md5.hexdigest()


def store_proteins_descriptions(pgdb, fastafn, fastamd5, tsvfn, header, fastadelim, genefield):
    if not fastafn:
        prots = {}
        for psm in tsvreader.generate_split_tsv_lines(tsvfn, header):
            prots.update({x: 1 for x in
                             tsvreader.get_proteins_from_psm(psm)})
        prots = [(protein,) for protein in prots.keys()]
        pgdb.store_proteins(prots)
    else:
        prots, seqs, desc, evids, ensgs, symbols = fastareader.get_proteins_for_db(
            fastafn, fastadelim, genefield)
        pgdb.store_fasta(fastafn, fastamd5, prots, evids, seqs, desc, ensgs, symbols)
    return set([x[0] for x in prots])


def store_psm_protein_relations(fn, header, pgdb, proteins, specfncol):
    """Reads PSMs from file, extracts their proteins and peptides and passes
    them to a database backend in chunks.
    """
    # TODO do we need an OrderedDict or is regular dict enough?
    # Sorting for psm_id useful?
    allpsms = OrderedDict()
    last_id, psmids_to_store = None, set()
    store_soon = False
    for psm in tsvreader.generate_split_tsv_lines(fn, header):
        psm_id, prots = tsvreader.get_pepproteins(psm, specfncol)
        # TODO can this be removed permanently? 
        # Filter proteins to only include those that match the protein 
        # accessions in fasta so we get the correct names, filter out the badly annotated peptides
        # prots = [x for x in prots if x in proteins]
        try:
            # In case the PSMs are presented unrolled
            allpsms[psm_id].extend(prots)
        except KeyError:
            allpsms[psm_id] = prots
        if len(psmids_to_store) > DB_STORE_CHUNK:
            store_soon = True
        if store_soon and last_id != psm_id:
            pgdb.store_peptides_proteins(allpsms, psmids_to_store)
            store_soon = False
            psmids_to_store = set()
        psmids_to_store.add(psm_id)
        last_id = psm_id
    if len(psmids_to_store) > 0:
        pgdb.store_peptides_proteins(allpsms, psmids_to_store)
    pgdb.index_protein_peptides()
    return allpsms

def add_genes_to_psm_table(psms, pgdb):
    gpmap = pgdb.get_protein_gene_map()
    for psm in psms:
        outpsm = {x: y for x, y in psm.items()}
        proteins = tsvreader.get_proteins_from_psm(psm)
        outpsm[mzidtsvdata.HEADER_GENE] = ';'.join(get_genes(proteins, gpmap))
        symbols = get_symbols(proteins, gpmap)
        desc = get_descriptions(proteins, gpmap)
        outpsm[mzidtsvdata.HEADER_SYMBOL] = ';'.join(symbols)
        outpsm[mzidtsvdata.HEADER_DESCRIPTION] = ';'.join(desc)
        yield outpsm


def get_mapped(proteins, gpmap, outtype):
    # FIXME multiple vals for a protein?
    outvals = [gpmap[protein][outtype] for protein in proteins]
    outvals = OrderedDict([(val, 1) for val in outvals]).keys()
    if None in outvals:
        return ['NA']
    else:
        return outvals


def get_genes(proteins, gpmap):
    return get_mapped(proteins, gpmap, 'gene')


def get_descriptions(proteins, gpmap):
    descriptions = get_mapped(proteins, gpmap, 'desc')
    return [x.replace('\t', ' ') for x in descriptions]


def get_symbols(proteins, gpmap):
    return get_mapped(proteins, gpmap, 'symbol')


def get_all_proteins_from_unrolled_psm(rownr, pgdb):
    return pgdb.get_proteins_for_peptide([rownr])


# FIXME tsv
def generate_psms_with_proteingroups(psms, pgdb, specfncol, unroll):
    rownr = 0
    use_evi = pgdb.check_evidence_tables()
    all_protein_group_content = pgdb.get_all_psms_proteingroups(use_evi)
    protein = next(all_protein_group_content)
    for psm in psms:
        if unroll:
            psm_id = tsvreader.get_psm_id(psm, specfncol)
            lineproteins = get_all_proteins_from_unrolled_psm(psm_id, pgdb)
        else:
            lineproteins = tsvreader.get_proteins_from_psm(psm)
        proteins_in_groups = {}
        while protein[0] == rownr:
            try:
                proteins_in_groups[protein[
                    lookups.MASTER_INDEX]].append(protein)
            except KeyError:
                proteins_in_groups[protein[lookups.MASTER_INDEX]] = [protein]
            try:
                protein = next(all_protein_group_content)
            except StopIteration:
                protein = [-1]
        sorted_pgs = sort_protein_groups(proteins_in_groups, use_evi)
        psm_masters = []
        psm_pg_proteins = []
        for master, group in sorted(sorted_pgs.items()):
            psm_masters.append(master)
            psm_pg_proteins.append([protein[lookups.PROTEIN_ACC_INDEX]
                                    for protein in group])
        outpsm = {mzidtsvdata.HEADER_MASTER_PROT: ';'.join(psm_masters),
                  mzidtsvdata.HEADER_PG_CONTENT: ';'.join(
                      [','.join([y for y in x]) for x in psm_pg_proteins]),
                  mzidtsvdata.HEADER_PG_AMOUNT_PROTEIN_HITS: ';'.join(
                      count_protein_group_hits(lineproteins, psm_pg_proteins))
                  }
        outpsm.update(psm)
        rownr += 1
        yield outpsm


def count_protein_group_hits(lineproteins, groups):
    """Takes a list of protein accessions and a list of protein groups
    content from DB. Counts for each group in list how many proteins
    are found in lineproteins. Returns list of str amounts.
    """
    hits = []
    for group in groups:
        hits.append(0)
        for protein in lineproteins:
            if protein in group:
                hits[-1] += 1
    return [str(x) for x in hits]


def get_sortfnxs(evidence):
    sortfnxs = [sort_pgroup_peptides,
                sort_pgroup_psms,
                sort_pgroup_score,
                ]
    if evidence:
        sortfnxs.append(sort_evidence_score)
    sortfnxs.extend([sort_pgroup_coverage, sort_alphabet])
    return sortfnxs


def sort_to_get_master(pgroup, evidence):
    sortfnxs = get_sortfnxs(evidence)
    sorted_pg = sort_protein_group(pgroup, sortfnxs, 0)
    return {'master_id': sorted_pg[0][lookups.MASTER_INDEX],
            'protein_acc': sorted_pg[0][lookups.PROTEIN_ACC_INDEX]}


def sort_protein_groups(pgroups, evidence):
    """Gets a protein groups containing dict pgroups, for each master (key)
    there is a list of protein tuples that comprise a proteingroup. This loops
    the groups and returns a sorted group. Assumes master is already sorted
    as the proper master"""
    sortfnxs = get_sortfnxs(evidence)
    pgroups_out = {}
    for master, pgroup in pgroups.items():
        sorted_pg = sort_protein_group(pgroup, sortfnxs, 0)
        pgroups_out[master] = sorted_pg
    return pgroups_out


def sort_protein_group(pgroup, sortfunctions, sortfunc_index):
    """Recursive function that sorts protein group by a number of sorting
    functions."""
    pgroup_out = []
    subgroups = sortfunctions[sortfunc_index](pgroup)
    sortfunc_index += 1
    for subgroup in subgroups:
        if len(subgroup) > 1 and sortfunc_index < len(sortfunctions):
            pgroup_out.extend(sort_protein_group(subgroup,
                                                 sortfunctions,
                                                 sortfunc_index))
        else:
            pgroup_out.extend(subgroup)
    return pgroup_out


def sort_amounts(proteins, sort_index):
    """Generic function for sorting peptides and psms. Assumes a higher
    number is better for what is passed at sort_index position in protein."""
    amounts = {}
    for protein in proteins:
        amount_x_for_protein = protein[sort_index]
        try:
            amounts[amount_x_for_protein].append(protein)
        except KeyError:
            amounts[amount_x_for_protein] = [protein]
    return [v for k, v in sorted(amounts.items(), reverse=True)]


def sort_pgroup_peptides(proteins):
    return sort_amounts(proteins, lookups.PEPTIDE_COUNT_INDEX)


def sort_pgroup_psms(proteins):
    return sort_amounts(proteins, lookups.PSM_COUNT_INDEX)


def sort_pgroup_score(proteins):
    return sort_amounts(proteins, lookups.PROTEIN_SCORE_INDEX)


def sort_pgroup_coverage(proteins):
    return sort_amounts(proteins, lookups.COVERAGE_INDEX)


def sort_evidence_score(proteins):
    return sort_amounts(proteins, lookups.EVIDENCE_LVL_INDEX)


def sort_alphabet(proteins):
    return [sorted(proteins, key=lambda x: x[2])]


def build_proteingroup_db(pgdb):
    build_master_db(pgdb)
    build_coverage(pgdb)
    build_content_db(pgdb)


def build_master_db(pgdb):
    psm_masters = OrderedDict()
    allmasters = {}
    allpsms, protpsms = {}, {}
    for psmid, protein in pgdb.get_all_psm_protein_relations():
        try:
            allpsms[psmid].add(protein)
        except KeyError:
            allpsms[psmid] = {protein}
        try:
            protpsms[protein].add(psmid)
        except KeyError:
            protpsms[protein] = {psmid}
    # PSMs with only one protein automatically make that protein a master
    # first extract all of them and their PSMs
    unipsms = {psm: list(prots)[0] for psm, prots in allpsms.items() if len(prots) == 1}
    for psm_id, master in unipsms.items():
        allmasters[master] = 1
        for grouppsm in protpsms[master]:
            try:
                psm_masters[grouppsm].add(master)
            except KeyError:
                psm_masters[grouppsm] = {master}
    # Add all explained PSMs to the filter so we dont explain them again,
    # They however are still in the graph (allpsms) and can get additional masters
    explained_psms = {x for x in psm_masters}

    def get_pp_graph(proteins, allpsms, protpsms):
        """When passed some proteins and lookups, this works out a graph of
        all protein/PSM connections"""
        graph = {x: set() for x in proteins}
        while True:
            nrprots, nrpsms = len(graph), sum([len(x) for x in graph.values()])
            for prot in list(graph.keys()): # list to avoid changing dict while iterating it
                if prot in protpsms:
                    graph[prot] = protpsms[prot]
                    for psm in graph[prot]:
                        if psm in allpsms:
                            graph.update({pr: set() for pr in allpsms[psm] if pr not in graph})
            if len(graph) == nrprots and sum([len(x) for x in graph.values()]) == nrpsms:
                break
        return graph

    # Use only PSMs that are not unique for the rest of the explaining
    # For each unexplained PSM, get the complete connection graph, fetch masters
    allpsms = {psm: prots for psm, prots in allpsms.items() if len(prots) > 1}
    for psm_id, proteins in allpsms.items():
        if psm_id in explained_psms:
            continue
        ppmap = get_pp_graph(proteins, allpsms, protpsms)
        explained_psms.update({y for x in ppmap.values() for y in x})
        masters = get_masters(ppmap)
        masterprots = {}
        for psm, psmmasters in masters.items():
            for master in psmmasters:
                masterprots[master] = 1
        allmasters.update(masterprots)
        for master, psm in pgdb.get_psms_for_proteins(list(masterprots.keys())):
            try:
                psm_masters[psm].add(master)
            except KeyError:
                psm_masters[psm] = {master}
    print('Collected {0} masters, {1} PSM-master mappings'.format(
        len(allmasters), len(psm_masters)))
    pgdb.store_masters(allmasters, psm_masters)


def process_pgroup_candidates(candidates, protein_psm_map):
    prepgroup = {}
    for candidate in candidates:
        master, psm_id, prot_id, seq, score, evid, cov = candidate
        protpsm_unit = (psm_id, float(score), evid, cov)
        try:
            prepgroup[prot_id][seq].add(protpsm_unit)
        except KeyError:
            try:
                prepgroup[prot_id][seq] = {protpsm_unit}
            except KeyError:
                prepgroup[prot_id] = {seq: {protpsm_unit}}
    pgroup = filter_proteins_with_missing_psms(prepgroup, protein_psm_map)
    return get_protein_group_content(pgroup, master)


def build_content_db(pgdb):
    protein_psms = {}
    for prot, psm in pgdb.get_protein_psm_records():
        try:
            protein_psms[prot].add(psm)
        except KeyError:
            protein_psms[prot] = {psm}
    use_evi = pgdb.check_evidence_tables()
    pg_candidates = pgdb.get_protein_group_candidates()
    pre_protein_group = [next(pg_candidates)]
    lastmaster = pre_protein_group[0][0]
    protein_groups, new_masters = [], {}
    for protein_candidate in pg_candidates:
        if protein_candidate[0] != lastmaster:
            pgroup = process_pgroup_candidates(pre_protein_group, protein_psms)
            new_master = sort_to_get_master(pgroup, use_evi)
            new_masters[new_master['master_id']] = new_master['protein_acc']
            protein_groups.extend(pgroup)
            lastmaster, pre_protein_group = protein_candidate[0], []
        pre_protein_group.append(protein_candidate)
    pgroup = process_pgroup_candidates(pre_protein_group, protein_psms)
    new_master = sort_to_get_master(pgroup, use_evi)
    new_masters[new_master['master_id']] = new_master['protein_acc']
    protein_groups.extend(pgroup)
    protein_groups = [[pg[2], pg[1], pg[3], pg[4], pg[5]]
                      for pg in protein_groups]
    new_masters = ((acc, mid) for mid, acc in new_masters.items())
    pgdb.update_master_proteins(new_masters)
    pgdb.store_protein_group_content(protein_groups)
    pgdb.index_protein_group_content()


def add_protein_psm_to_pre_proteingroup(prepgmap, protein, pepseq,
                                        psm_id, score, evid, cover):
    protpsm_unit = (psm_id, float(score), evid, cover)
    try:
        prepgmap[protein][pepseq].add(protpsm_unit)
    except KeyError:
        try:
            prepgmap[protein][pepseq] = {protpsm_unit}
        except KeyError:
            prepgmap[protein] = {pepseq: {protpsm_unit}}
    return prepgmap


def filter_proteins_with_missing_psms(proteins, allprotein_psms):
    filtered_protein_map = {}
    for protein, protein_psms in proteins.items():
        protein_masterpsms = {psm[0] for peptide in protein_psms.values()
                              for psm in peptide}
        if allprotein_psms[protein].difference(protein_masterpsms):
            continue
        else:
            filtered_protein_map[protein] = protein_psms
    return filtered_protein_map


def build_coverage(pgdb):
    coverage = {}
    for acc, seq, psm_id, psmseq in pgdb.get_all_proteins_psms_seq():
        try:
            coverage[acc]['seq'] = seq
        except KeyError:
            coverage[acc] = {'seq': seq, 'psms': []}
        coverage[acc]['psms'].append(psmseq)
    pgdb.store_coverage(generate_coverage(coverage))


def get_masters(ppgraph):
    """From a protein-peptide graph dictionary (keys proteins,
    values peptides), return master proteins aka those which
    have no proteins whose peptides are supersets of them.
    If shared master proteins are found, report only the first,
    we will sort the whole proteingroup later anyway. In that
    case, the master reported here may be temporary."""
    masters = {}
    for protein, peps in ppgraph.items():
        ismaster = True
        multimaster = set()
        for subprotein, subpeps in ppgraph.items():
            if protein == subprotein:
                continue
            if peps.issubset(subpeps):
                if peps.union(subpeps) > peps:
                    ismaster = False
                    break
                elif peps.intersection(subpeps) == peps:
                    multimaster.update({protein, subprotein})
        if not ismaster:
            continue
        elif multimaster:
            premaster = sorted(list(multimaster))[0]
        else:
            premaster = protein
        for pep in peps:
            try:
                masters[pep].add(premaster)
            except KeyError:
                masters[pep] = {premaster}
    return masters


def generate_coverage(seqinfo):
    """From a dict containing protein accessions and sequences/PSM sequences,
    this function returns a generator that calculates coverages for each
    protein and returns the accession and coverage percentage.
    Coverage is done by finding peptides in the protein seq using seq.index
    and marking the range. May be slow."""
    for acc, protinfo in seqinfo.items():
        coverage_aa_indices = set()
        seq = protinfo['seq']
        for psmseq in protinfo['psms']:
            psmseq = tsvreader.strip_modifications(psmseq)
            # FIXME try block is for problems with coverage, see if it is
            # needed
            try:
                start = seq.index(psmseq)
            except:
                print('CANNOT FIND PSM seq {0} in seq {1} '
                      'for acc {2}'.format(psmseq, seq, acc))
            coverage_aa_indices.update(range(start, start + len(psmseq)))
        yield (acc, len(coverage_aa_indices) / len(seq))


def get_protein_group_content(pgmap, master):
    """For each master protein, we generate the protein group proteins
    complete with sequences, psm_ids and scores. Master proteins are included
    in this group.

    Returns a list of [protein, master, pep_hits, psm_hits, protein_score],
    which is ready to enter the DB table.
    """
    # first item (0) is only a placeholder so the lookup.INDEX things get the
    # correct number. Would be nice with a solution, but the INDEXes were
    # originally made for mzidtsv protein group adding.
    pg_content = [[0, master, protein, len(peptides), len([psm for pgpsms in
                                                           peptides.values()
                                                           for psm in pgpsms]),
                   sum([psm[1] for pgpsms in peptides.values()
                        for psm in pgpsms]),  # score
                   next(iter(next(iter(peptides.values()))))[3],  # coverage
                   next(iter(next(iter(peptides.values()))))[2],  # evid level
                   ]
                  for protein, peptides in pgmap.items()]
    return pg_content


def generate_psms_quanted(quantdb, shiftrows, psms, isob_header, isobaric,
        precursor, min_prec_purity):
    """Takes dbfn and connects, gets quants for each line in tsvfn, sorts
    them in line by using keys in quantheader list."""
    allquants, sqlfields = quantdb.select_all_psm_quants(shiftrows, isobaric, precursor)
    quant = next(allquants)
    for rownr, psm in enumerate(psms):
        rownr += shiftrows
        outpsm = {x: y for x, y in psm.items()}
        if precursor:
            pquant = [quant[sqlfields['precursor']], quant[sqlfields['fwhm']]]
            pquant = ['NA' if x is None else x for x in pquant]
            outpsm.update({
                mzidtsvdata.HEADER_PRECURSOR_QUANT: str(pquant[0]),
                mzidtsvdata.HEADER_PRECURSOR_FWHM: str(pquant[1]),
                })
        if isobaric:
            isoquants, pif = {}, None
            while quant[0] == rownr:
                isoquants.update({quant[sqlfields['isochan']]:
                                  str(quant[sqlfields['isoquant']])})
                pif = quant[sqlfields['pif']]
                try:
                    quant = next(allquants)
                except StopIteration:
                    # last PSM, break from while loop or it is not yielded at all
                    break
            outpsm.update(get_quant_NAs(isoquants, isob_header, pif, min_prec_purity))
        else:
            try:
                quant = next(allquants)
            except StopIteration:
                # last PSM, needs explicit yield/break or it will not be yielded
                yield outpsm
                break
        yield outpsm


def get_quant_NAs(quantdata, quantheader, purity, min_purity):
    """Takes quantdata in a dict and header with quantkeys
    (eg iTRAQ isotopes). Returns dict of quant intensities
    with missing keys set to NA."""
    comp_purity = purity if purity is not None else 0
    if comp_purity >= min_purity:
        out = {k: quantdata.get(k, 'NA') for k in quantheader}
    else:
        out = {k: 'NA' for k in quantheader}
    out[mzidtsvdata.HEADER_PREC_PURITY] = str(purity or 'NA')
    return out


def count_missed_cleavage(full_pepseq, count=0):
    '''Regex .*[KR][^P] matches until the end and checks if there is a final
    charachter so this will not match the tryptic residue'''
    pepseq = re.sub('[\+\-]\d*.\d*', '', full_pepseq)
    match = re.match('.*[KR][^P]', pepseq)
    if match:
        count += 1
        return count_missed_cleavage(match.group()[:-1], count)
    else:
        return count


def generate_psms_spectradata(lookup, shiftrows, psms, bioset, miscleav):
    psm_specdata = zip(enumerate(psms), lookup.get_exp_spectra_data_rows(shiftrows))
    for (row, psm), specdata in psm_specdata:
        row += shiftrows
        outpsm = {x: y for x, y in psm.items()}
        if row == int(specdata[0]):
            inj = str(specdata[3])
            imob = str(specdata[4])
            outpsm.update({mzidtsvdata.HEADER_RETENTION_TIME: str(specdata[2]),
                           mzidtsvdata.HEADER_INJECTION_TIME: inj if inj != 'None' else 'NA',
                           mzidtsvdata.HEADER_ION_MOB: imob if imob != 'None' else 'NA',
                           })
            if bioset:
                outpsm[mzidtsvdata.HEADER_SETNAME] = str(specdata[1])
        else:
            raise RuntimeError('PSM with row nr {} has no rownr in DB. '
                               'Current DB row is {}'.format(row, specdata[0]))
        if miscleav:
            outpsm[mzidtsvdata.HEADER_MISSED_CLEAVAGE] = count_missed_cleavage(outpsm[mzidtsvdata.HEADER_PEPTIDE])
        yield outpsm



from app.lookups.sqlite import (quant, searchspace, biosets, spectra, 
        prottable, psms, peptable)


def get_lookup(fn, lookuptype):
    lookupmap = {'spectra': spectra.SpectraDB,
                 'specquant': quant.QuantDB,
                 'psm': psms.PSMDB,
                 'searchspace': searchspace.SearchSpaceDB,
                 'peptidetable': peptable.PepTableProteinCentricDB,
                 'peptidegenecentrictable': peptable.PepTableGeneCentricDB,
                 'peptidetableplain': peptable.PepTablePlainDB,
                 'prottable': prottable.ProtTableDB,
                 'plainprottable': prottable.ProtTableNoGroupNoMapDB,
                 'genetable': prottable.GeneTableDB,
                 'associdtable': prottable.GeneTableAssocIDsDB,
                 }
    return lookupmap[lookuptype](fn)


def create_new_lookup(fn, lookuptype):
    with open(fn, 'w'):
        pass
    return get_lookup(fn, lookuptype)



import re

from app.dataformats import mzidtsv as header
from app.actions.percolator.filters import sequence_to_filterseqs


def filter_psms_conf(psms, confkey, conflvl, lower_is_better):
    threshold = float(conflvl)
    for psm in psms:
        try:
            confval = float(psm[confkey])
        except (TypeError, ValueError):
            pass
        else:
            lower = confval < threshold
            if lower == lower_is_better:
                yield psm


def filter_psms_remove_set(psms, setnames):
    biosets = set(setnames)
    for psm in psms:
        if psm[header.HEADER_SETNAME] not in biosets:
            yield psm


def remove_duplicate_psms(psms, speccol, scancol, seqcol):
    '''Removes PSMs that are duplicate, i.e. have identical file/scan/sequence
    combinations'''
    known_seqids = set() 
    for psm in psms:
        seqid = f'{psm[speccol]}_{psm[scancol]}_{psm[seqcol]}'
        if seqid not in known_seqids:
            known_seqids.add(seqid)
            yield psm


def filter_known_searchspace(psms, lookup, seqcol, ntermwildcards, deamidation):
    """Yields peptides from generator as long as their sequence is not found in
    known search space dict. Useful for excluding peptides that are found in
    e.g. ENSEMBL or similar"""
    for psm in psms:
        seq_is_known = False
        seq = re.sub('[^A-Z]', '', psm[seqcol].upper())
        for seq in  sequence_to_filterseqs(seq, deamidation):
            if lookup.check_seq_exists(seq, ntermwildcards):
                seq_is_known = True
                break
        if not seq_is_known:
            yield psm


def filter_whole_proteins(psms, lookup, seqcol, deamidation, minpeplen, enforce_tryp):
    for psm in psms:
        seq_matches_protein = False
        seq = re.sub('[^A-Z]', '', psm[seqcol].upper())
        filterseqs = sequence_to_filterseqs(seq, deamidation)
        found_prots = lookup.get_proteins_from_peps(filterseqs, minpeplen)
        for pepseq, proteins in found_prots.items():
            for prot_id, pos, protseq in proteins:
                if pepseq in protseq.replace('L', 'I'):
                    if enforce_tryp and (pos == 0 or not set(
                            [pepseq[-1],
                             protseq[pos - 1]]).difference(['K', 'R'])):
                        # pepseq is tryptic on both ends, or
                        # pepseq is an N-term peptide),
                        # matches to protein seq so remove
                        seq_matches_protein = True
                        break
                    elif not enforce_tryp:
                        seq_matches_protein = True
                        break
        if not seq_matches_protein:
            yield psm


def match_sequence(psms, lookup, seqcol, matchfield, ntermwildcards, deamidation, fullprot_minlen, enforce_tryp):
    '''Yields peptides with a new field, which contains accession(s) to 
    protein(s) with matching peptide sequence in the passed storeseq lookup
    '''
    for psm in psms:
        outpsm = {k: v for k,v in psm.items()}
        seq_is_known = False
        seq = re.sub('[^A-Z]', '', psm[seqcol].upper())
        filterseqs = sequence_to_filterseqs(seq, deamidation)
        found_prots = lookup.get_proteins_matching_peptides(filterseqs, ntermwildcards, fullprot_minlen)
        if fullprot_minlen:
            protids = set()
            for pepseq, proteins in found_prots.items():
                for prot_id, pos, protseq in proteins:
                    if pepseq in protseq.replace('L', 'I'):
                        if enforce_tryp and (pos == 0 or not set(
                                [pepseq[-1],
                                 protseq[pos - 1]]).difference(['K', 'R'])):
                            # pepseq is tryptic on both ends, or
                            # pepseq is an N-term peptide),
                            # matches to protein seq so remove
                            protids.add(prot_id)
                        elif not enforce_tryp:
                            protids.add(prot_id)
            outpsm[matchfield] = ';'.join(protids) if protids else 'No match'
        else:
            outpsm[matchfield] = ';'.join(found_prots) if found_prots else 'No match'
        yield outpsm



import os
import sys
from math import log
from statistics import median, StatisticsError
from collections import OrderedDict

from app.dataformats import prottable as prottabledata
from app.dataformats import mzidtsv as psmh
from app.readers import tsv as reader


ISOQUANTRATIO_FEAT_ACC = '##isoquant_target_acc##'


def get_isobaric_ratios(psmfn, psmheader, channels, denom_channels, sweep,
        report_intensity, summarize_by, min_int, targetfeats, target_acc_field, accessioncol,
        totalprot_feats, totalp_field, totalp_pepfield, logintensities, mediannormalize,
        medianfactors, keep_na_psms, split_multi_entries=False):
    """Main function to calculate ratios for PSMs, peptides, proteins, genes.
    Can do simple ratios, median-of-ratios, median-centering, log2, etc
    """
    outratios = get_psmratios(psmfn, psmheader, channels, denom_channels,
            sweep, report_intensity, summarize_by, min_int, accessioncol, logintensities,
            keep_na_psms, split_multi_entries)
    # at this point, outratios look like:
    # [{ch1: 123, ch2: 456, ISOQUANTRATIO_FEAT_ACC: ENSG1244}, ]
    if accessioncol:
        if mediannormalize:
            # median center ratios before proceeding to output
            # with or without factor report (do not need it if using another
            # table since that is implied to have been analysed elsewhere anyway)
            if medianfactors:
                report_factors = False
            else:
                report_factors = os.path.basename(psmfn)
            outratios = mediancenter_ratios(outratios, channels, logintensities, medianfactors, report_factors)
        if totalprot_feats:
            # immediately output
            return totalproteome_normalization(outratios, targetfeats, 
                    target_acc_field, channels, totalprot_feats, totalp_field,
                    totalp_pepfield, logintensities)
        if targetfeats:
            # Output to the target feat table (e.g. existing peptable, protein table etc)
            outratios = {x.pop(ISOQUANTRATIO_FEAT_ACC): x for x in outratios}
            return output_to_targetfeats(targetfeats, target_acc_field, outratios, channels)
        else:
            # generate new table with accessions if no existing ID table is specified
            return ({(k if not k == ISOQUANTRATIO_FEAT_ACC else accessioncol): v
                     for k, v in ratio.items()} for ratio in outratios)
    else: 
        return paste_to_psmtable(psmfn, psmheader, outratios)


def totalproteome_normalization(outratios, targetfeats, acc_field, channels, totalprot,
        totalp_field_tp, totalp_field_target, logratios):
    """Normalizes PTM-peptides (mostly phospho usually) to their respective protein
    ratios, by dividing or subtracting (log data) intensity ratios with the 
    total proteome intensity ratios. Peptides which match to multiple proteins
    are set to NA
    """
    outratios = {x.pop(ISOQUANTRATIO_FEAT_ACC): x for x in outratios}
    totalprot = {x[totalp_field_tp]: x for x in totalprot}
    for feat in targetfeats:
        outfeat = {k: v for k,v in feat.items()}
        try:
            quants = outratios[feat[acc_field]]
        except KeyError:
            quants = {ch: 'NA' for ch in channels}
            outfeat.update(quants)
            yield outfeat
        else:
            totalp_acc = feat[totalp_field_target]
            norm_q = {k: v for k,v in quants.items()}
            if ';' in totalp_acc or totalp_acc not in totalprot:
                norm_q.update({ch: 'NA' for ch in channels})
            elif logratios:
                norm_q.update({ch: norm_q[ch] - float(totalprot[totalp_acc][ch])
                    if quants[ch] != 'NA' and totalprot[totalp_acc][ch] != 'NA'
                    else 'NA' for ch in channels})
            else:
                norm_q.update({ch: norm_q[ch] / float(totalprot[totalp_acc][ch])
                    if quants[ch] != 'NA' and totalprot[totalp_acc][ch] != 'NA'
                    else 'NA' for ch in channels})
            outfeat.update(norm_q)
            yield outfeat
        # FIXMNE correct medians? Get 0.0. for all, suspicious

        
def mediancenter_ratios(ratios, channels, logratios, factor_table, report_factors):
    if factor_table:
        factor_ratios = [[float(x[ch]) if x[ch] != 'NA' else 'NA'
            for ch in channels] for x in factor_table]
    else:
        factor_ratios = [[x[ch] for ch in channels] for x in ratios]
    ch_medians = get_medians(channels, factor_ratios, basename_report=report_factors)
    for quant in ratios:
        if logratios:
            quant.update({ch: quant[ch] - ch_medians[ch]
                if quant[ch] != 'NA' else 'NA' for ch in channels})
        else:
            quant.update({ch: quant[ch] / ch_medians[ch]
                if quant[ch] != 'NA' else 'NA' for ch in channels})
        yield quant


def get_psmratios(psmfn, header, channels, denom_channels, sweep, report_intensity, 
        summarize_by, min_int, acc_col, logintensities, keep_na_psms, split_multi_entries):
    allfeats, feat_order, psmratios = {}, OrderedDict(), []
    for psm in reader.generate_split_tsv_lines(psmfn, header):
        # remove uninformative psms when adding to features
        # TODO the check for is-not-a-peptide can be removed but there are some usecases
        # for which it is convenient, when adding information to the peptide
        # sequence (e.g. PTM data). When having fully functional PTM
        # data storage/analysis in msstitch, we can possibly remove it
        if acc_col and not split_multi_entries:
            rm_multi_map = ';' in psm[acc_col]
            accessions = [psm[acc_col]]
        elif acc_col:
            rm_multi_map = False
            accessions = psm[acc_col].split(';')
        if acc_col and (psm[acc_col] == '' or 
                (acc_col != psmh.HEADER_PEPTIDE and rm_multi_map) or
                not {psm[q] for q in channels}.difference({'NA', None, False, ''})
                ):
            continue
        ratios = calc_psm_ratios_or_int(psm, channels, denom_channels, sweep, 
                report_intensity, min_int, logintensities)
        if acc_col and not keep_na_psms and any((ratios[ix] == 'NA' for ix, q in enumerate(channels))):
            continue
        elif acc_col:
            for accession in accessions:
                try:
                    allfeats[accession].append(ratios)
                except KeyError:
                    allfeats[accession] = [ratios]
                feat_order[accession] = 1
        else:
            psmquant = {ch: str(ratios[ix]) if ratios[ix] != 'NA' else 'NA'
                        for ix, ch in enumerate(channels)}
            psmquant[ISOQUANTRATIO_FEAT_ACC] = False
            psmratios.append(psmquant)
    if not acc_col:
        return psmratios
    else:
        outfeatures = []
        for feat in feat_order.keys():
            quants = allfeats[feat]
            outfeature = {ISOQUANTRATIO_FEAT_ACC: feat}
            if summarize_by == 'median':
                outfeature.update(get_medians(channels, quants))
            elif summarize_by == 'average':
                outfeature.update(summarize_by_averages(channels, quants))
            outfeature.update(get_no_psms(channels, quants))
            outfeatures.append(outfeature)
    return outfeatures


def paste_to_psmtable(psmfn, header, ratios):
    # loop psms in psmtable, paste the outratios in memory
    for psm, ratio in zip(reader.generate_split_tsv_lines(psmfn, header), ratios):
        ratio.pop(ISOQUANTRATIO_FEAT_ACC)
        ratio = {'ratio_{}'.format(ch): val for ch, val in ratio.items()}
        psm.update(ratio)
        yield psm


def output_to_targetfeats(targetfeats, acc_field, featratios, channels):
    #loop prottable, add ratios from dict, acc = key
    for feat in targetfeats:
        try:
            quants = featratios[feat[acc_field]]
        except KeyError:
            quants = {ch: 'NA' for ch in channels}
            quants.update({get_no_psms_field(ch): 'NA' for ch in channels})
        feat.update(quants)
        yield feat


def calc_psm_ratios_or_int(psm, channels, denom_channels, sweep, report_intensity,
        min_intensity, logintensities):
    # set values below min_intensity to NA
    if logintensities:
        psm_intensity = {ch: log(float(psm[ch]), 2)
                         if psm[ch] != 'NA' and float(psm[ch]) > min_intensity
                         else 'NA' for ch in channels}
    else:
        psm_intensity = {ch: float(psm[ch])
                         if psm[ch] != 'NA' and float(psm[ch]) > min_intensity
                         else 'NA' for ch in channels}
    if denom_channels:
        denomvalues = [psm_intensity[ch] for ch in denom_channels
                       if psm_intensity[ch] != 'NA']
    elif sweep:
        try:
            denomvalues = [median([x for x in psm_intensity.values() if x != 'NA'])]
        except StatisticsError:
            # Empty PSM errors on median call
            denomvalues = []
    elif report_intensity:
        # Just report intensity
        return [psm_intensity[ch] if psm_intensity[ch] != 'NA' else 'NA' for ch in channels]
    if (not logintensities and sum(denomvalues) == 0) or len(denomvalues) == 0:
        return ['NA'] * len(channels)
    # TODO add median instead of average?
    # TODO can we use means of logged values or is that not correct? DEqMS does use it
    denom = sum(denomvalues) / len(denomvalues)
    if logintensities:
        return [psm_intensity[ch] - denom
                if psm_intensity[ch] != 'NA' else 'NA' for ch in channels]
    else:
        return [psm_intensity[ch] / denom
                if psm_intensity[ch] != 'NA' else 'NA' for ch in channels]


def summarize_by_averages(channels, ratios):
    ch_avgs = {}
    for ix, channel in enumerate(channels):
        vals = [x[ix] for x in ratios if x[ix] != 'NA']
        try:
            ch_avgs[channel] = sum(vals) / len(vals)
        except ZeroDivisionError:
            # channel is empty
            ch_avgs[channel] = 'NA'
    return ch_avgs


def get_medians(channels, ratios, basename_report=False):
    ch_medians = {}
    for ix, channel in enumerate(channels):
        try:
            ch_medians[channel] = median([x[ix] for x in ratios
                                          if x[ix] != 'NA'])
        except StatisticsError:
            # channel is empty, common in protein quant but not in normalizing
            ch_medians[channel] = 'NA'
    if basename_report:
        reporttext = '{}\n'.format('\n'.join(['{}\t{}'.format(ch, ch_medians[ch]) for ch in channels]))
        with open('normalization_factors_{}'.format(basename_report), 'w') as fp:
            fp.write(reporttext)
    return ch_medians


def get_no_psms_field(quantfield):
    return '{}{}'.format(quantfield, prottabledata.HEADER_NO_PSMS_SUFFIX)


def get_no_psms(channels, ratios):
    ch_nopsms = {}
    for ix, channel in enumerate(channels):
        fieldname = get_no_psms_field(channel)
        ch_nopsms[fieldname] = len([x[ix] for x in ratios if x[ix] != 'NA'])
    ch_nopsms[prottabledata.HEADER_NO_FULLQ_PSMS] = len([x for x in ratios if 'NA' not in x])
    return ch_nopsms



from app.drivers import base
from app.drivers.options import percolator_options

from app.actions.percolator import split
from app.actions.percolator import filters


class FilterSequences(base.PercolatorDriver):
    """This class processes multiple percolator runs from fractions and
    filters out first peptides that are found in a specified searchspace. Then
    it keeps the remaining best scoring unique peptides."""
    outsuffix = '_filtered.xml'
    command = 'filterperco'
    lookuptype = 'searchspace'
    commandhelp = ('Filters out peptides that match to a protein in a FASTA '
                   'file. Needs a lookup to match peptides to quickly.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['fullprotein', 'deamidate', 
            'minlength', 'lookupfn', 'forcetryp', 'insourcefrag'],
            percolator_options))

    def set_features(self):
        if self.fullprotein:
            self.features = {
                'peptide': filters.filter_whole_proteins(self.allpeps,
                                                             self.lookup,
                                                             'pep',
                                                             self.ns,
                                                             self.deamidate,
                                                             self.minlength,
                                                             self.forcetryp),
                'psm': filters.filter_whole_proteins(self.allpsms,
                                                         self.lookup,
                                                         'psm',
                                                         self.ns,
                                                         self.deamidate,
                                                         self.minlength,
                                                         self.forcetryp)
            }
        else:
            self.features = {
                'peptide': filters.filter_known_searchspace(self.allpeps,
                                                                'pep',
                                                                self.lookup,
                                                                self.ns,
                                                                self.insourcefrag,
                                                                self.deamidate),
                'psm': filters.filter_known_searchspace(self.allpsms,
                                                            'psm',
                                                            self.lookup,
                                                            self.ns,
                                                            self.insourcefrag,
                                                            self.deamidate),
            }


class SplitDriver(base.PercolatorDriver):
    outfile = None

    def run(self):
        self.set_filter_types()
        for filter_type, suffix in self.filter_types:
            self.prepare()
            self.set_features(filter_type)
            self.outsuffix = suffix
            self.write()

    def set_options(self):
        """Since splitdriver splits into multiple files we cannot set an
        output file"""
        super().set_options()
        del(self.options['outfile'])

    def set_features(self, filter_type):
        """Calls splitter to split percolator output into target/decoy
        elements.
        Writes two new xml files with features. Currently only psms and
        peptides. Proteins not here, since one cannot do protein inference
        before having merged and remapped multifraction data anyway.
        """
        elements_to_split = {'psm': self.allpsms, 'peptide': self.allpeps}
        self.features = self.splitfunc(elements_to_split, self.ns, filter_type)


class FilterDuplicatePeptideDriver(base.PercolatorDriver):
    '''This class removes duplicate peptides from a percolator run, in case
    there have been changes to the output after percolator scoring (e.g. removal
    of modifications).
    '''
    outsuffix = '_pepdedup.xml'
    command = 'dedupperco'
    commandhelp = 'Filters out duplicate peptides (by sequence) from percolator output if these are there'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['deduppsms'], percolator_options))

    def set_features(self):
        if self.deduppsms:
            psmfilter = filters.dedup_psms
        else:
            psmfilter = filters.passthrough
        self.features = {
                'peptide': filters.filter_best_peptide(self.allpeps, self.ns),
                'psm': psmfilter(self.allpsms, self.ns),
                }



class SplitProteinDriver(SplitDriver):
    """Using --protheaders "novel_ENSP" "ENSP|variant_"  will result in two output files,
    the first containing all PSM/peptides containing at least one mapping with a header novel_ENSP,
    and the second containing all PSMs/peptides containing at least one mapping with header variant_,
    but no PSMs/peptides which map to ENSP headers.
    """
    command = 'splitperco'
    commandhelp = ('Splits input XML into multiple files depending based on '
                   'the protein headers specified. Each header class gets '
                   'its own output file')

    def set_filter_types(self):
        maxdigits = len(str(len(self.protheaders)))
        self.filter_types = [(headers, '_h{i:0{dig}d}.xml'.format(
            i=ix, dig=maxdigits))
            for ix, headers in enumerate(self.protheaders)]

    def set_features(self, filter_type):
        self.splitfunc = split.split_protein_header_id_type
        super().set_features(filter_type)

    def set_options(self):
        super().set_options()
        options = self.define_options(['protheaders'], percolator_options)
        self.options.update(options)



from tests.integration import basetests

import os
import sqlite3
import json
from Bio import SeqIO
from lxml import etree


class SearchspaceLookup(basetests.BaseTest):
    suffix = ''
    infilename = 'proteins.fasta'

    def setUp(self):
        super().setUp()
        self.infile = os.path.join(self.basefixdir, self.infilename)


class TestTrypsinize(SearchspaceLookup):
    suffix = '_tryp.fa'
    command = 'trypsinize'

    def run_case(self, minlen, cutproline, miscleav, ntermloss, ignorestopcodons):
        options = ['-o', self.resultfn]
        if minlen:
            options.extend(['--minlen', str(minlen)])
        if cutproline:
            options.extend(['--cutproline'])
            seqtype = 'proline_cuts'
        if ntermloss and miscleav and ignorestopcodons:
            options.extend(['--nterm-meth-loss', '--miscleav', str(miscleav), '--ignore-stop-codons'])
            seqtype = 'fully_tryptic_minlen_ntermloss_ignorestop'
        elif ntermloss and miscleav:
            options.extend(['--nterm-meth-loss', '--miscleav', str(miscleav)])
            seqtype = 'fully_tryptic_ntermloss'
        elif miscleav:
            options.extend(['--miscleav', str(miscleav)])
            seqtype = 'miscleav'
        self.run_command(options)
        with open(os.path.join(self.basefixdir, 'peptides_trypsinized.json')) as fp:
            tryp_sequences = json.load(fp)
        foundseqs = {}
        for rec in SeqIO.parse(self.resultfn, 'fasta'):
            self.assertEqual(tryp_sequences[seqtype][str(rec.seq)], rec.id)
            foundseqs[str(rec.seq)] = rec.id
            if minlen:
                self.assertGreaterEqual(len(str(rec.seq)), minlen)
        for pep, prot in tryp_sequences[seqtype].items():
            if not minlen or len(pep) >= minlen:
                self.assertEqual(prot, foundseqs[pep])

    def test_fullytryptic_minlen_ntermloss_stopcodons(self):
        self.run_case(8, False, 1, True, True)

    def test_prolinecut(self):
        self.run_case(False, True, False, False, False)

    def test_miss_cleavage(self):
        self.run_case(False, False, 1, False, False)

    def test_nterm_miss_cleavage(self):
        '''Specific code path for this, so test it'''
        self.run_case(False, False, 1, True, False)


class TestDecoyFa(SearchspaceLookup):
    command = 'makedecoy'
    infilename = 'twoproteins.fasta'

    def run_check(self, options):
        self.resultfn = os.path.join(self.workdir, 'decoy.fa')
        options.extend(['-o', self.resultfn])
        self.run_command(options)

    def run_without_predb(self, options):
        self.run_check(options)

    def run_with_existing_db(self, options):
        self.copy_db_to_workdir('decoycheck.sqlite', 'decoycheck.sqlite')
        options.extend(['--dbfile', 'decoycheck.sqlite'])
        self.run_check(options)

    def check_seqs(self, checkfile, dbcheck=False, keep_maxscrambled=False):
        targetfa = SeqIO.index(self.infile[0], 'fasta')
        checkfa = SeqIO.index(os.path.join(self.fixdir, checkfile), 'fasta')
        resfa = SeqIO.index(self.resultfn, 'fasta')
        ttryps = {}
        for tid, t in targetfa.items():
            ttryps[tid] = [y for x in t.seq.split('K') for y in x.split('R')]
        allttryps = (y for x in ttryps.values() for y in x)
        # Are all seqs from result in expected?
        for seqid, seq in resfa.items():
            check = checkfa[seqid]
            trypseqs = [y for x in seq.seq.split('K') for y in x.split('R')]
            trypchecks = [y for x in check.seq.split('K') for y in x.split('R')]
            try:
                self.assertEqual(seq.seq, check.seq)
            except AssertionError:
                if dbcheck:
                    tseqs = ttryps[seq.id.replace('decoy_', '')]
                    for resseq, tseq in zip(trypseqs, tseqs):
                        if resseq in allttryps:
                            self.assertEqual(resseq[-1], tseq[-1])
                            self.assertEqual(len(resseq), len(tseq))
                    self.assertEqual(set(seq.seq), set(check.seq))
                    self.assertEqual(len(seq), len(check))
                else:
                    raise
        # Are all seqs from check in result?
        for seqid, seq in checkfa.items():
            self.assertEqual(seq.seq, resfa[seqid].seq)

    def test_tryprev_predb_keeptargets(self):
        '''Tests making decoy where target hits are validated against DB, NOT removed if
        matching target despite shuffling'''
        self.run_with_existing_db(['--scramble', 'tryp_rev', '--maxshuffle', '10', '--keep-target'])
        self.check_seqs('decoy_tryprev_tcheck_keeptarget_twoproteins.fasta', dbcheck=True, keep_maxscrambled=True)

    def test_tryprev_yesdb(self):
        '''Tests making decoy where target hits are validated against DB, removed if
        matching target despite shuffling'''
        self.run_without_predb(['--scramble', 'tryp_rev'])
        self.check_seqs('decoy_tryprev_targetcheck_twoproteins.fasta', dbcheck=True)

    def test_tryprev_yesdb_minlen(self):
        '''Tests making decoy where target hits are validated against DB, removed if
        matching target despite shuffling and also removed if len < 5'''
        self.run_without_predb(['--scramble', 'tryp_rev', '--minlen', '5'])
        self.check_seqs('decoy_tryprev_minlen_twoproteins.fasta', dbcheck=True)

    def test_tryprev_ignore_db(self):
        '''Tests decoy where target hits are kept and no DB is checked'''
        self.run_without_predb(['--scramble', 'tryp_rev', '--ignore-target-hits'])
        self.check_seqs('decoy_tryprev_twoproteins.fasta', dbcheck=True)

    def test_protrev(self):
        '''Tests making decoy using protein-reverse, no target checking is done here'''
        self.run_without_predb(['--scramble', 'prot_rev'])
        self.check_seqs('decoy_twoproteins.fasta')

    def test_pretryp_predb(self):
        '''Tests making decoy where sequences are trypsinized before hand, and supplying
        an existing SQLite DB for matching target. Persistent (despite shuffling) target-matching 
        sequences will be removed'''
        options = ['--scramble', 'tryp_rev', '--notrypsin']
        self.infilename = 'trypsinized_twoproteins.fasta'
        self.infile = [os.path.join(self.fixdir, self.infilename)]
        self.run_with_existing_db(options)
        self.check_seqs('decoy_tryprev_pretryp_twoproteins.fasta', True)

    def test_pretryp_predb_keeptarget(self):
        '''Tests making decoy where sequences are trypsinized before hand, and supplying
        an existing SQLite DB for matching target. Persistent (despite shuffling) target-matching 
        sequences will not be removed'''
        options = ['--scramble', 'tryp_rev', '--notrypsin', '--keep-target']
        self.infilename = 'trypsinized_twoproteins.fasta'
        self.infile = [os.path.join(self.fixdir, self.infilename)]
        self.run_with_existing_db(options)
        self.check_seqs('decoy_tryprev_pretryp_keeptarget_twoproteins.fasta', True)


class TestTrypticLookup(SearchspaceLookup):
    command = 'storeseq'

    def query_db_assert(self, options=None, seqtype=None):
        if options is None:
            options = []
        with open(os.path.join(self.basefixdir, 'peptides_trypsinized.json')) as fp:
            tryp_sequences = json.load(fp)
        if seqtype is not None:
            sequences = tryp_sequences[seqtype]
        else:
            sequences = tryp_sequences['fully_tryptic']
        self.run_command(options)
        db = sqlite3.connect(self.resultfn)
        seqs_in_db = []
        for seq in sequences:
            seq = seq.replace('L', 'I')
            sql = ('SELECT EXISTS(SELECT seqs FROM known_searchspace WHERE '
                       'seqs=? LIMIT 1)')
            seqs_in_db.append((seq, db.execute(sql, (seq,)).fetchone()[0] == 1))
        [self.assertTrue(x[1]) for x in seqs_in_db]
        seqcount_db = db.execute('SELECT COUNT(*) FROM known_searchspace').fetchone()[0]
        db.close()
        self.assertEqual(len(sequences), seqcount_db)

    def run_without_db(self, options=None, seqtype=None):
        self.resultfn = os.path.join(self.workdir, 'mslookup_db.sqlite')
        self.query_db_assert(options, seqtype)

    def run_with_existing_db(self, options=None, seqtype=None):
        if options is None:
            options = []
        self.resultfn = os.path.join(self.workdir, 'seqspace.db')
        options.extend(['--dbfile', self.resultfn])
        self.copy_db_to_workdir('spectra_lookup.sqlite')
        self.query_db_assert(options, seqtype)

    def test_cutproline_nodb(self):
        self.run_without_db(['--cutproline'], 'proline_cuts')

    def test_cutproline_yesdb(self):
        self.run_with_existing_db(['--cutproline'], 'proline_cuts')

    def test_ntermwildcards_nodb(self):
        self.run_without_db(['--insourcefrag'], 'insourcefrag')

    def test_ntermwildcards_yes_db(self):
        self.run_with_existing_db(['--insourcefrag'], 'insourcefrag')

    def test_ntermmethloss_nodb(self):
        self.run_without_db(['--nterm-meth-loss', '--miscleav', '1'], 'fully_tryptic_ntermloss')

    def test_ntermmethloss_yesdb(self):
        self.run_with_existing_db(['--nterm-meth-loss', '--miscleav', '1'], 'fully_tryptic_ntermloss')

    def test_noflags_no_db(self):
        self.run_without_db()

    def test_noflags_yes_db(self):
        self.run_with_existing_db()


class TestWholeProteinSeqLookup(SearchspaceLookup):
    command = 'storeseq'

    def query_db_assert(self, options):
        with open(os.path.join(self.basefixdir, 'allpeptides_proteins.json')) as fp:
            sequences = [x.replace('L', 'I') for x in json.load(fp)]
        options.extend(['--fullprotein', '--minlen', '6'])
        self.run_command(options)
        db = sqlite3.connect(self.resultfn)
        seqs_in_db = []
        sql = ('SELECT EXISTS(SELECT seq FROM protein_peptides WHERE '
               'seq=? LIMIT 1)')
        for seq in sequences:
            seqs_in_db.append((seq, db.execute(sql, (seq,)).fetchone()[0] == 1))
        seqcount_db = db.execute('SELECT COUNT() FROM protein_peptides').fetchone()[0]
        db.close()
        [self.assertTrue(x[1]) for x in seqs_in_db]
        self.assertEqual(len(sequences), seqcount_db)

    def test_without_db(self):
        self.resultfn = os.path.join(self.workdir, 'mslookup_db.sqlite')
        self.query_db_assert([])

    def test_with_existing_db(self):
        self.resultfn = os.path.join(self.workdir, 'seqspace.db')
        options = ['--dbfile', self.resultfn]
        self.copy_db_to_workdir('spectra_lookup.sqlite')
        self.query_db_assert(options)


class TestPepProtMapLookup(SearchspaceLookup):
    command = 'storeseq'

    def run_without_db(self, options, seqtype):
        dbfn = 'mslookup_db.sqlite'
        self.resultfn = os.path.join(self.workdir, dbfn)
        with open(os.path.join(self.basefixdir, 'peptides_trypsinized.json')) as fp:
            sequences = [x.replace('L', 'I') for x in json.load(fp)[seqtype]]
        self.run_command(['--map-accessions', *options])
        db = sqlite3.connect(dbfn)
        seqs_in_db = set()
        sql = ('SELECT EXISTS(SELECT seq FROM protein_peptides WHERE '
               'seq=? LIMIT 1)')
        for seq in sequences:
            seqs_in_db.add((seq, db.execute(sql, (seq,)).fetchone()[0] == 1))
        print(seqs_in_db)
        [self.assertTrue(x[1]) for x in seqs_in_db]
        seqcount_db = db.execute('SELECT COUNT(*) FROM protein_peptides').fetchone()[0]
        db.close()
        self.assertEqual(len(sequences), seqcount_db)

    def test_fulltryp_nodb(self):
        self.run_without_db([], 'fully_tryptic')

    def test_ntermloss_miscleav_nodb(self):
        self.run_without_db(['--nterm-meth-loss', '--miscleav', '1'], 'fully_tryptic_ntermloss')

    def test_insourcefrag_nodb(self):
        self.run_without_db(['--insourcefrag'], 'insourcefrag')

    def test_cutproline_nodb(self):
        self.run_without_db(['--cutproline'], 'proline_cuts')


class SpectraLookup(basetests.MSLookupTest):
    command = 'storespectra'

    def setUp(self):
        super().setUp()
        self.infile = os.path.join(self.basefixdir, self.infilename)

    def get_std_options(self):
        return [self.executable, self.command, '--spectra', self.infile]

    def check_spectra(self, bsets, ionmob=False):
        sql = ('SELECT mf.mzmlfilename, bs.set_name, s.scan_sid, s.charge, '
               's.mz, s.retention_time, s.spectra_id '
               '{} '
               'FROM mzml AS s '
               '{} '
               'JOIN mzmlfiles AS mf USING(mzmlfile_id) '
               'JOIN biosets AS bs USING(set_id)')
        if ionmob:
            ionsql = (', imob.ion_mobility', ' JOIN ionmob AS imob USING(spectra_id) ')
        else:
            ionsql = (', iit.ion_injection_time', ' JOIN ioninjtime AS iit USING(spectra_id) ')
        sql = sql.format(*ionsql)
        specrecs = {}
        for rec in self.get_values_from_db(self.resultfn, sql):
            specrecs[rec[2]] = {'fn': rec[0], 'bs': rec[1], 'charge': rec[3],
                                'mz': rec[4], 'rt': rec[5], 
                                'sid': rec[6],
                                'ion_something': rec[7],
                                }
        for scannr, spec in self.get_spectra_mzml([self.infile], bsets, ionmob):
            self.assertEqual(spec, specrecs[scannr])

    def get_spectra_mzml(self, infiles, bsets, ionmob):
        def get_cvparam_value(parent, name, ns):
            return [x.attrib['value'] for x in
                    parent.findall('{%s}cvParam' % ns['xmlns'])
                    if x.attrib['name'] == name]

        def multifind(elements, parent, ns):
            element = elements.pop(-1)
            if len(elements) > 0:
                parent = multifind(elements, parent, ns)
            return parent.find('{%s}%s' % (ns['xmlns'], element))

        for mzfnid, (bset, infile) in enumerate(zip(bsets, infiles)):
            ns = self.get_xml_namespace(infile)
            for ac, spectrum in etree.iterparse(
                    infile, tag='{%s}spectrum' % ns['xmlns']):
                if get_cvparam_value(spectrum, 'ms level', ns)[0] != "2":
                    continue
                scansid = spectrum.attrib['id']
                scan = multifind(['scanList', 'scan'], spectrum, ns)
                precursor = multifind(['precursorList', 'precursor',
                                       'selectedIonList', 'selectedIon'],
                                      spectrum, ns)
                rtval = get_cvparam_value(scan, 'scan start time', ns)[0]
                if ionmob:
                    ion = get_cvparam_value(scan, 'inverse reduced ion mobility', ns)[0]
                    rt = float(rtval) / 60
                else:
                    ion = get_cvparam_value(scan, 'ion injection time', ns)[0]
                    rt = float(rtval)
                charge = get_cvparam_value(precursor, 'charge state', ns)
                charge = charge[0] if len(charge) else False
                mz = get_cvparam_value(precursor, 'selected ion m/z', ns)[0]
                exp_data = {'fn': os.path.basename(infile), 'bs': bset,
                            'charge': int(charge),
                            'sid': '{}_{}'.format(mzfnid + 1, scansid),
                            'mz': float(mz), 'rt': round(rt, 12), 'ion_something': round(float(ion), 12)}
                yield scansid, exp_data


class TestDDAThermoSpectraLookup(SpectraLookup):
    infilename = 'few_spectra.mzML'

    def test_spectra_newdb(self):
        setnames = ['Set1']
        options = ['--setnames']
        options.extend(setnames)
        self.run_command(options)
        self.check_spectra(setnames)

    def test_spectra_newdb_filename(self):
        setnames = ['Set1']
        outfn = 'testresult.sql'
        self.resultfn = os.path.join(self.workdir, outfn)
        options = ['-o', outfn, '--setnames']
        options.extend(setnames)
        self.run_command(options)
        self.check_spectra(setnames)


class TestDDATIMSSpectraLookup(SpectraLookup):
    infilename = 'few_spec_timstof.mzML'

    def test_spectra_newdb(self):
        setnames = ['Set1']
        options = ['--setnames']
        options.extend(setnames)
        self.run_command(options)
        self.check_spectra(setnames, ionmob=True)


class TestSpecQuantLookup(basetests.MSLookupTest):
    command = 'storequant'
    infilename = ''
    isoinfilename = 'few_spectra.consXML'
    isoinfilename2 = 'few_spec_timstof.consXML'
    krinfilename = 'few_spectra.kr'
    dinoinfilename = 'few_spectra.dino'
    base_db_fn = 'spectra_lookup.sqlite'

    def setUp(self):
        super().setUp()
        self.isoinfile = os.path.join(self.fixdir, self.isoinfilename)
        self.isoinfile2 = os.path.join(self.fixdir, self.isoinfilename2)
        self.krfile = os.path.join(self.fixdir, self.krinfilename)
        self.dinofile = os.path.join(self.fixdir, self.dinoinfilename)
        self.fakespfn = os.path.join(self.workdir, 'few_spectra.mzML')
        self.fakespfn2 = os.path.join(self.workdir, 'set2.mzML')
        self.fakespfn3 = os.path.join(self.workdir, 'few_spec_timstof.mzML')

    def get_std_options(self):
        return [self.executable, self.command]

    def get_quantch_map(self):
        qmap_xml = {}
        for ac, map_el in etree.iterparse(self.isoinfile, tag='map'):
            channeldata = map_el.attrib
            qmap_xml[channeldata['id']] = channeldata['label']
        return qmap_xml

    def check_quantmap(self):
        qmap_xml = self.get_quantch_map()
        qmap_xml = [qmap_xml[x]for x in sorted(qmap_xml.keys())]
        sql = 'SELECT channel_name FROM isobaric_channels'
        for xml_channel, db_rec in zip(qmap_xml,
                                       self.get_values_from_db(self.resultfn,
                                                               sql)):
            self.assertEqual(xml_channel, db_rec[0])

    def check_quantification(self):
        sql = ('SELECT iq.intensity, ic.channel_name, pif.pif FROM isobaric_quant '
               'AS iq JOIN isobaric_channels AS ic USING(channel_id) '
               'LEFT OUTER JOIN precursor_ion_fraction AS pif USING(spectra_id)')
        qch_map = self.get_quantch_map()
        dbtmt = self.get_values_from_db(self.resultfn, sql)
        for ac, xml_quant in etree.iterparse(self.isoinfile,
                                             tag='consensusElement'):
            for element in xml_quant.findall('.//element'):
                qval, qchan, pif = next(dbtmt)
                self.assertEqual(qch_map[element.attrib['map']], qchan)
                self.assertEqual(float(element.attrib['it']), qval)
            xmlpif = xml_quant.xpath('./UserParam[@name="precursor_purity"]')
            if len(xmlpif):
                self.assertEqual(float(xmlpif[0].attrib['value']), pif)


    def test_isoquant_mixtypes(self):
        options = ['--isobaric', self.isoinfile, self.isoinfile2, '--spectra', self.fakespfn, self.fakespfn3]
        self.run_command(options)
        self.check_quantmap()
        self.check_quantification()

    def test_isoquant(self):
        options = ['--isobaric', self.isoinfile, '--spectra', self.fakespfn]
        self.run_command(options)
        self.check_quantmap()
        self.check_quantification()

    def test_dinosaur(self):
        options = ['--dinosaur', self.dinofile, self.dinofile, '--rttol', '5', 
                '--mztol', '20', '--mztoltype', 'ppm', '--spectra', self.fakespfn, 
                self.fakespfn2]
        self.run_command(options)
        self.check_ms1_feats_stored(self.dinofile, 'dino', 'sum')

    def test_dinosaur_apex(self):
        options = ['--dinosaur', self.dinofile, self.dinofile, '--rttol', '5',
                '--mztol', '20', '--apex', '--mztoltype', 'ppm', '--spectra', self.fakespfn,
                self.fakespfn2]
        self.run_command(options)
        self.check_ms1_feats_stored(self.dinofile, 'dino', 'apex')

    def test_kronik(self):
        options = ['--kronik', self.krfile, self.krfile, '--rttol', '5',
                '--mztol', '20', '--mztoltype', 'ppm', '--spectra',
                self.fakespfn, self.fakespfn2]
        self.run_command(options)
        self.check_ms1_feats_stored(self.krfile, 'kr', 'sum')

    def test_kronik_apex(self):
        options = ['--kronik', self.krfile, self.krfile, '--rttol', '5',
                '--mztol', '20', '--apex', '--mztoltype', 'ppm', '--spectra',
                self.fakespfn, self.fakespfn2]
        self.run_command(options)
        self.check_ms1_feats_stored(self.krfile, 'kr', 'apex')

    def check_ms1_feats_stored(self, ms1file, feattype, intkey):
        PROTON_MASS = 1.0072
        sql = ('SELECT count(*) FROM ms1_quant LEFT OUTER JOIN ms1_fwhm USING(feature_id)')
        recs = self.get_values_from_db(self.resultfn, sql)
        sql = ('SELECT * FROM ms1_quant LEFT OUTER JOIN ms1_fwhm USING(feature_id)')
        feats = {1: {}, 2: {}}
        for recid, fnid, rt, mz, charge, val, fwhm in self.get_values_from_db(
                self.resultfn, sql):
            try:
                feats[fnid][rt][mz][charge] = (val, fwhm)
            except KeyError:
                try:
                    feats[fnid][rt][mz] = {charge: (val, fwhm)}
                except KeyError:
                    feats[fnid][rt] = {mz: {charge: (val, fwhm)}}
        fields = {
                'dino': {
                    'charge': 'charge',
                    'rt': 'rtApex',
                    'intensity': {'sum': 'intensitySum', 
                        'apex': 'intensityApex'}[intkey]
                    },
                'kr': {
                    'charge': 'Charge',
                    'rt': 'Best RTime',
                    'intensity': {'sum': 'Summed Intensity', 
                        'apex': 'Best Intensity'}[intkey]
                    },
                }[feattype]
        with open(ms1file) as fp:
            header = next(fp).strip().split('\t')
            for line in fp:
                line = {k: v for k, v in zip(header, line.strip('\n').split('\t'))}
                charge = float(line[fields['charge']])
                if feattype == 'kr':
                    line['fwhm'] = None
                    mz = (float(line['Monoisotopic Mass']) + charge * PROTON_MASS) / charge
                else:
                    mz = float(line['mz'])
                resval = feats[1][round(float(line[fields['rt']]), 12)][mz][charge]
                fwhm = float(line['fwhm']) if line['fwhm'] else None
                self.assertEqual((float(line[fields['intensity']]), fwhm), resval)
        # Now check we have a feature for all scans
        sql = """SELECT m.scan_sid, ma.feature_id FROM mzml AS m
        JOIN mzmlfiles AS mzf USING(mzmlfile_id)
        JOIN biosets AS bs USING(set_id)
        LEFT OUTER JOIN ms1_align AS ma USING(spectra_id)
        WHERE bs.set_name != 'Set3'
        """
        exemptscans = {'dino': [10229], 'kr': [10148, 10229]}
        exemptscans = [f'controllerType=0 controllerNumber=1 scan={x}' for x 
                in exemptscans[feattype]]
        for scan, featid in self.get_values_from_db(self.resultfn, sql):
            # some scans do not have a aligned MS1 value
            if scan not in exemptscans:
                self.assertFalse(featid == None)
            else:
                self.assertTrue(featid == None)



import os
import re
from math import log

from tests.integration import basetests


class TestPeptideMerge(basetests.MergeTest):
    infilename = 'target_peptides.tsv'

    def test_proteincentric(self):
        self.options.extend(['--pepcolpattern', 'peptide PEP'])
        self.run_command(self.options)
        sql = ('SELECT ps.sequence, p.psm_id, prot.protein_acc, pd.description, '
               'g.gene_acc, aid.assoc_id, pc.coverage, pseq.sequence '
               'FROM peptide_sequences AS ps '
               'JOIN psms AS p USING(pep_id) '
               'JOIN psm_protein_groups USING(psm_id) '
               'JOIN protein_group_master AS pm USING(master_id) '
               'JOIN proteins AS prot USING(pacc_id) '
               'JOIN protein_seq AS pseq USING(protein_acc) '
               'LEFT OUTER JOIN prot_desc AS pd USING(pacc_id) '
               'LEFT OUTER JOIN ensg_proteins AS egp USING(pacc_id) '
               'LEFT OUTER JOIN genes AS g USING(gene_id) '
               'LEFT OUTER JOIN genename_proteins AS gnp USING(pacc_id) '
               'LEFT OUTER JOIN associated_ids AS aid USING(gn_id) '
               'JOIN protein_coverage AS pc USING(protein_acc) ')
        self.check_iso_and_peptide_relations(sql, proteincentric=True)

    def test_genecentric(self):
        self.options.append('--genecentric')
        self.options.extend(['--pepcolpattern', 'peptide PEP'])
        self.options.extend(['--flrcolpattern', 'q-value'])
        self.run_command(self.options)
        sql = ('SELECT ps.sequence, p.psm_id, "NA", pd.description, '
               'g.gene_acc, aid.assoc_id, "NA", "NA" '
               'FROM peptide_sequences AS ps '
               'JOIN psms AS p USING(pep_id) '
               'JOIN protein_psm USING(psm_id) '
               'JOIN proteins AS prot USING(protein_acc) '
               'LEFT OUTER JOIN prot_desc AS pd USING(pacc_id) '
               'LEFT OUTER JOIN ensg_proteins AS egp USING(pacc_id) '
               'LEFT OUTER JOIN genes AS g USING(gene_id) '
               'LEFT OUTER JOIN genename_proteins AS gnp USING(pacc_id) '
               'LEFT OUTER JOIN associated_ids AS aid USING(gn_id) '
               'JOIN protein_coverage AS pc USING(protein_acc) '
               )
        self.check_iso_and_peptide_relations(sql, flr=True)

    def test_nogroups(self):
        self.options.extend(['--pepcolpattern', 'peptide PEP'])
        self.options.append('--no-group-annotation')
        self.run_command(self.options)
        sql = ('SELECT ps.sequence, p.psm_id, prot.protein_acc, "NA", "NA", "NA", "NA", pseq.sequence '
               'FROM peptide_sequences AS ps '
               'JOIN psms AS p USING(pep_id) '
               'JOIN protein_psm USING(psm_id) '
               'JOIN proteins AS prot USING(protein_acc) '
               'JOIN protein_seq AS pseq USING(protein_acc) '
               )
        self.check_iso_and_peptide_relations(sql, proteincentric=True, nogroup=True)

    def check_iso_and_peptide_relations(self, sql, proteincentric=False, nogroup=False, flr=False):
        valsql = ('SELECT ps.sequence, bs.set_name, '
               'ppq.quant, pf.fdr, pp.pep, flr.flr '
               'FROM peptide_sequences AS ps '
               'JOIN biosets AS bs '
               'JOIN peptide_precur_quanted AS ppq USING(pep_id) '
               'LEFT OUTER JOIN ptm_flr AS flr USING(pep_id) '
               'JOIN peptide_fdr AS pf USING(pep_id) '
               'JOIN peptide_pep AS pp USING(pep_id) ')
        fields = ['MS1 area (highest of all PSMs)', 'q-value', 'PEP']
        if flr:
            fields += ['PTM FLR']
        self.check_build_values(valsql, fields, 'Peptide sequence')
        isosql = ('SELECT ps.sequence, bs.set_name, ch.channel_name, '
               'iq.quantvalue, iq.amount_psms, fqp.amount_psms '
               'FROM peptide_sequences AS ps '
               'JOIN biosets AS bs '
               'JOIN peptide_iso_quanted AS iq USING(pep_id) '
               'JOIN pepquant_channels AS ch USING(channel_id) '
               'JOIN peptide_iso_fullpsms AS fqp USING(pep_id) '
               )
        self.check_built_isobaric(isosql, 'Peptide sequence')
        expected, psm_id = {}, None
        for rec in self.get_values_from_db(self.dbfile, sql):
            rec = [rec[0]] + ['NA' if x is None else x for x in rec[1:]]
            try:
                expected[rec[0]]['psms'].add(rec[1])
            except KeyError:
                expected[rec[0]] = {'psms': set([rec[1]]),
                                    'pgroups': set([rec[2]]),
                                    'descriptions': set([rec[3]]),
                                    'genes': set([rec[4]]),
                                    'assoc': set([rec[5]]),
                                    'cover': set([rec[6]]),
                                    'pseqs': {rec[2]: rec[7]},
                                    }
            else:
                expected[rec[0]]['pgroups'].add(rec[2])
                expected[rec[0]]['descriptions'].add(rec[3])
                expected[rec[0]]['genes'].add(rec[4])
                expected[rec[0]]['assoc'].add(rec[5])
                expected[rec[0]]['cover'].add(rec[6])
                expected[rec[0]]['pseqs'][rec[2]] = rec[7]
        for line in self.tsv_generator(self.resultfn):
            pepseq = line['Peptide sequence']
            barepep = re.sub('[^A-Z]', '', pepseq)
            self.assertEqual(line['Bare peptide'], barepep)
            if proteincentric or nogroup:
                proteins = line['Protein(s)'].split(';')
                for pacc, ss in zip(proteins, line['Protein start/stop site(s)'].split(';')):
                    start, stop = [int(x) for x in ss.split('-')]
                    ix = expected[pepseq]['pseqs'][pacc].index(barepep)
                    self.assertEqual(start, ix + 1)
                    self.assertEqual(stop, ix + len(barepep))
                
            if proteincentric:
                self.assertEqual(set(line['Protein(s)'].split(';')),
                                 expected[pepseq]['pgroups'])
                if not nogroup:
                    self.assertEqual(
                        set(line['Description(s)'].split(';')),
                        set([y for x in
                             expected[pepseq]['descriptions']
                             for y in x.split(';')]))
                    rescovs = sorted(line['Coverage(s)'].split(';'))
                    expcovs = sorted(expected[pepseq]['cover'])
                    for rescov, expcov in zip(rescovs, expcovs):
                        self.assertAlmostEqual(float(rescov), expcov)
            if not nogroup:
                self.assertEqual(set(line['Gene ID(s)'].split(';')),
                                 expected[pepseq]['genes'])
                self.assertEqual(set(line['Gene name(s)'].split(';')),
                                 expected[pepseq]['assoc'])


class TestProteinMerge(basetests.MergeTest):

    def proteins(self, cutoff=False):
        self.infilename = 'proteins.txt'
        if cutoff:
            self.options.extend(['--mergecutoff', str(cutoff)])
        self.run_command(self.options)
        sql = ('SELECT p.protein_acc, bs.set_name, pf.fdr, pp.quant FROM proteins AS p '
               'JOIN biosets AS bs '
               'JOIN protein_fdr AS pf USING(pacc_id) '
               'JOIN protein_precur_quanted AS pp USING(pacc_id) '
               )
        self.check_build_values(sql, ['q-value', 'MS1 precursor area'],
                'Protein ID', cutoff)
        sql = ('SELECT p.protein_acc, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms, fqp.amount_psms, pf.fdr '
               'FROM proteins AS p '
               'JOIN biosets AS bs '
               'JOIN protein_iso_quanted AS pi USING(pacc_id) '
               'JOIN protein_iso_fullpsms AS fqp USING(pacc_id) '
               'JOIN protquant_channels AS pc USING(channel_id) '
               'JOIN protein_fdr AS pf WHERE pf.prottable_id=pc.prottable_id '
               'AND pf.pacc_id=p.pacc_id'
               )
        self.check_built_isobaric(sql, 'Protein ID', cutoff=cutoff)
        sql = """
        SELECT p.protein_acc,GROUP_CONCAT(g.gene_acc),GROUP_CONCAT(aid.assoc_id),pd.description,pcov.coverage
        FROM protein_group_master AS pgm 
        INNER JOIN proteins AS p ON p.pacc_id=pgm.pacc_id 
        LEFT OUTER JOIN genename_proteins AS gnp ON gnp.pacc_id=p.pacc_id
        LEFT OUTER JOIN associated_ids AS aid ON aid.gn_id=gnp.gn_id
        LEFT OUTER JOIN ensg_proteins AS egp ON egp.pacc_id=p.pacc_id
        LEFT OUTER JOIN genes AS g ON g.gene_id=egp.gene_id
        LEFT OUTER JOIN prot_desc AS pd USING(pacc_id) 
        JOIN protein_coverage AS pcov ON pcov.protein_acc=p.protein_acc
        GROUP BY pgm.pacc_id
        """
        psm_sql = """
        SELECT x.protein_acc, pp.psm_id, ps.sequence
        FROM protein_group_master AS pgm
        INNER JOIN proteins AS x ON x.pacc_id=pgm.pacc_id
        JOIN protein_psm AS pp ON pp.protein_acc=x.protein_acc
        JOIN psms USING(psm_id)
        JOIN mzml USING(spectra_id)
        JOIN mzmlfiles USING(mzmlfile_id)
        JOIN biosets AS bs USING(set_id)
        JOIN peptide_sequences AS ps USING(pep_id)
        WHERE bs.set_name='Set1'
        """
        self.check_protein_data('proteincentric', sql, psm_sql)

    def test_proteins(self):
        self.proteins()

    def test_mergecutoff(self):
        # FIXME need worse peptides, all FDR is 0.0, cutoff not tested!!
        # test is in check_built_isobaric
        self.proteins(0.0001)

    def test_nopsmnrs(self):
        """Given input tables with NO amount psm information, output a nice gene
        table without those columns"""
        isosql = """
        SELECT g.gene_acc, bs.set_name, pc.channel_name,
        pi.quantvalue, pi.amount_psms FROM genes AS g
        JOIN biosets AS bs
        JOIN gene_iso_quanted AS pi USING(gene_id)
        JOIN genequant_channels AS pc USING(channel_id)
        """
        self.ensgcentric('ensg_nopsms.txt', isosql, nopsms=True)

    def test_ensgcentric(self):
        isosql = """
        SELECT g.gene_acc, bs.set_name, pc.channel_name,
        pi.quantvalue, pi.amount_psms, fqp.amount_psms FROM genes AS g
        JOIN biosets AS bs
        JOIN gene_iso_quanted AS pi USING(gene_id)
        JOIN gene_iso_fullpsms AS fqp USING(gene_id)
        JOIN genequant_channels AS pc USING(channel_id)
        """
        self.ensgcentric('ensg.txt', isosql)

    def ensgcentric(self, infile, isosql, nopsms=False):
        self.infilename = infile
        self.run_command(self.options)
        sql = ('SELECT g.gene_acc, bs.set_name, gf.fdr, gp.quant '
               'FROM genes AS g '
               'JOIN biosets AS bs '
               'JOIN gene_fdr AS gf USING(gene_id) '
               'JOIN gene_precur_quanted AS gp USING(gene_id) '
               )
        self.check_build_values(sql, ['q-value', 'MS1 precursor area'],
                'Gene ID')
        self.check_built_isobaric(isosql, 'Gene ID', nopsms)
        sql = """
        SELECT g.gene_acc, GROUP_CONCAT(aid.assoc_id),GROUP_CONCAT(p.protein_acc),pd.description
        FROM genes AS g
        INNER JOIN ensg_proteins AS egp ON egp.gene_id=g.gene_id
        LEFT OUTER JOIN genename_proteins AS gnp ON gnp.pacc_id=egp.pacc_id
        JOIN associated_ids AS aid ON aid.gn_id=gnp.gn_id
        JOIN proteins AS p ON p.pacc_id=gnp.pacc_id
        JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id
        GROUP BY g.gene_acc
        """
        psm_sql = """
        SELECT x.gene_acc, pp.psm_id, ps.sequence
        FROM genes AS x
        INNER JOIN ensg_proteins AS egp ON egp.gene_id=x.gene_id
        INNER JOIN proteins AS p ON p.pacc_id=egp.pacc_id
        JOIN protein_psm AS pp ON pp.protein_acc=p.protein_acc
        JOIN psms USING(psm_id)
        JOIN mzml USING(spectra_id)
        JOIN mzmlfiles USING(mzmlfile_id)
        JOIN biosets AS bs USING(set_id)
        JOIN peptide_sequences AS ps USING(pep_id)
        WHERE bs.set_name='Set1'
        """
        self.check_protein_data('genecentric', sql, psm_sql)

    def test_genenamecentric(self):
        self.infilename = 'genenames.txt'
        self.run_command(self.options)
        sql = ('SELECT ai.assoc_id, bs.set_name, gf.fdr, gp.quant '
               'FROM associated_ids AS ai '
               'JOIN biosets AS bs '
               'JOIN assoc_fdr AS gf USING(gn_id) '
               'JOIN assoc_precur_quanted AS gp USING(gn_id) '
               )
        self.check_build_values(sql, ['q-value', 'MS1 precursor area'], 
                'Gene Name')
        sql = ('SELECT ai.assoc_id, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms, fqp.amount_psms '
               'FROM associated_ids AS ai '
               'JOIN biosets AS bs '
               'JOIN assoc_iso_quanted AS pi USING(gn_id) '
               'JOIN assoc_iso_fullpsms AS fqp USING(gn_id) '
               'JOIN genequant_channels AS pc USING(channel_id) '
               )
        self.check_built_isobaric(sql, 'Gene Name')
        sql = """
        SELECT aid.assoc_id, GROUP_CONCAT(g.gene_acc), GROUP_CONCAT(p.protein_acc),pd.description
        FROM associated_ids AS aid
        INNER JOIN genename_proteins AS gnp ON gnp.gn_id=aid.gn_id
        LEFT OUTER JOIN ensg_proteins AS egp ON egp.pacc_id=gnp.pacc_id
        JOIN genes AS g ON g.gene_id=egp.gene_id
        JOIN proteins AS p ON p.pacc_id=gnp.pacc_id
        JOIN prot_desc AS pd ON pd.pacc_id=p.pacc_id
        GROUP BY aid.assoc_id
        """
        psm_sql = """
        SELECT x.assoc_id, pp.psm_id, ps.sequence
        FROM associated_ids AS x
        INNER JOIN genename_proteins AS gnp ON gnp.gn_id=x.gn_id
        INNER JOIN proteins AS p ON p.pacc_id=gnp.pacc_id
        JOIN protein_psm AS pp ON pp.protein_acc=p.protein_acc
        JOIN psms USING(psm_id)
        JOIN mzml USING(spectra_id)
        JOIN mzmlfiles USING(mzmlfile_id)
        JOIN biosets AS bs USING(set_id)
        JOIN peptide_sequences AS ps USING(pep_id)
        WHERE bs.set_name='Set1'
        """
        self.check_protein_data('assoccentric', sql, psm_sql) 

    def check_peptide_relations(self, sql):
        expected, psm_id = {}, None
        for rec in self.get_values_from_db(self.dbfile, sql):
            rec = [rec[0]] + ['NA' if x is None else x for x in rec[1:]]
            try:
                expected[rec[0]]['psms'].add(rec[1])
            except KeyError:
                expected[rec[0]] = {'psms': set([rec[1]]),
                                    'pgroups': set([rec[2]]),
                                    'descriptions': set([rec[3]]),
                                    'genes': set([rec[4]]),
                                    'assoc': set([rec[5]]),
                                    'cover': set([str(rec[6])]),
                                    }
            else:
                expected[rec[0]]['pgroups'].add(rec[2])
                expected[rec[0]]['descriptions'].add(rec[3])
                expected[rec[0]]['genes'].add(rec[4])
                expected[rec[0]]['assoc'].add(rec[5])
                expected[rec[0]]['cover'].add(str(rec[6]))
        for line in self.tsv_generator(self.resultfn):
            self.assertEqual(set(line['Protein(s)'].split(';')),
                             expected[line['Peptide sequence']]['pgroups'])
            self.assertEqual(
                set(line['Description(s)'].split(';')),
                set([y for x in
                     expected[line['Peptide sequence']]['descriptions']
                     for y in x.split(';')]))
            self.assertEqual(set(line['Gene ID(s)'].split(';')),
                             expected[line['Peptide sequence']]['genes'])
            self.assertEqual(set(line['Gene name(s)'].split(';')),
                             expected[line['Peptide sequence']]['assoc'])
            self.assertEqual(set([line['Coverage(s)']]),
                             expected[line['Peptide sequence']]['cover'])
